import os
import sys

from .DominantColors import DominantColors

dir = sys.argv[1]
m = {}

for f in sorted(os.listdir(dir)):
    if f.lower().endswith(".jpg") or f.lower().endswith(".jpg"):
        try:
            d = DominantColors(os.path.join(dir, f))
            calc = d.get_dominant_colors()
            print(f, "light:", calc[2])
            continue

            for fuzzy in range(20):
                if DominantColors.contains_color(calc, (255, 217, 100), fuzzy):
                    m[f] = fuzzy
                    print(f, fuzzy)
                    break
            else:
                m[f] = -1
                print(f, "no match")

        except Exception:
            print("oops for " + f)
            raise
            # pass

# print "\n----results----"
# for fuzzy in xrange(10):
#    print fuzzy
#    for k, v in m.items():
#        if v == fuzzy:
#            print k
#
# print "no match:"
# for k, v in m.items():
#    if v == -1:
#        print k



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import abc
from typing import Callable, List, Optional

from variety.Util import Util

from .IVarietyPlugin import IVarietyPlugin


class DisplayModeData:
    """
    Contains the data for how to visualize a specific image.
    set_wallpaper_param - what do we send to the set_wallpaper script, affects OS background options
    imagemagick_cmd - optional, what command do we run over the image in order to resize it
    fixed_image_path - optional, if more complex logic needed, generate the image and give its path
    """

    def __init__(
        self,
        set_wallpaper_param: str,
        imagemagick_cmd: Optional[str] = None,
        fixed_image_path: Optional[str] = None,
    ):
        self.set_wallpaper_param = set_wallpaper_param
        self.imagemagick_cmd = imagemagick_cmd
        self.fixed_image_path = fixed_image_path


class DisplayMode:
    """
    Implements a display mode.
    Needs a unique id, title to show in the combobox, description to show below the combo when
    selected, and callable that implements the logic.
    The callable takes a file path and returns a DisplayModeData object.
    """

    def __init__(self, id: str, title: str, description: str, fn: Callable[[str], DisplayModeData]):
        self.id = id
        self.title = title
        self.description = description
        self.fn = fn


class StaticDisplayMode(DisplayMode):
    """
    A DisplayMode that does not care about the specific file, but implements a uses
    either a static ImageMagick command, or does no resizing at all and works simply via the
    set_wallpaper parameter.
    """

    def __init__(
        self,
        id: str,
        title: str,
        description: str,
        set_wallpaper_param: str,
        imagemagick_cmd: Optional[str] = None,
    ):
        def fn(filename: str):
            w, h = Util.get_primary_display_size()
            if imagemagick_cmd:
                final_cmd = imagemagick_cmd.replace("%W", str(w)).replace("%H", str(h))
            else:
                final_cmd = None
            return DisplayModeData(
                set_wallpaper_param=set_wallpaper_param, imagemagick_cmd=final_cmd
            )

        super().__init__(id, title, description, fn)


class IDisplayModesPlugin(IVarietyPlugin):
    @abc.abstractmethod
    def display_modes(self) -> List[DisplayMode]:
        """
        Return a list of DisplayModes
        """
        return []

    def order(self):
        """
        Display modes from IDisplayModesPlugins with lower order get listed before those with
        higher order
        """
        return 100



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import base64
import codecs
import datetime
import functools
import gettext
import hashlib
import json
import logging
import os
import random
import re
import shutil
import string
import subprocess
import sys
import threading
import time
import urllib.parse
from itertools import cycle

import bs4
import requests
from PIL import Image

from variety_lib import get_version

# fmt: off
import gi  # isort:skip
gi.require_version("GExiv2", "0.10")
gi.require_version("PangoCairo", "1.0")
gi.require_version('Gdk', '3.0')
from gi.repository import Gdk, GdkPixbuf, GExiv2, Gio, GLib, Pango  # isort:skip
# fmt: on


USER_AGENT = "Variety Wallpaper Changer " + get_version()

random.seed()
logger = logging.getLogger("variety")
gettext.textdomain("variety")


def _(text):
    """Returns the translated form of text."""
    if not text or not text.strip():
        return text
    return gettext.gettext(text)


def debounce(seconds):
    """ Decorator that will postpone a functions execution until after wait seconds
        have elapsed since the last time it was invoked. """

    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)

            try:
                debounced.t.cancel()
            except (AttributeError):
                pass
            debounced.t = threading.Timer(seconds, call_it)
            debounced.t.start()

        return debounced

    return decorator


class throttle(object):
    """
    Decorator that prevents a function from being called more than once every time period. Allows for a trailing call.

    To create a function that cannot be called more than once a minute:

        @throttle(seconds=1)
        def my_fun():
            pass
    """

    def __init__(self, seconds=0, trailing_call=False):
        """
        seconds - throttle interval in seconds
        trailing - if True, there will always be a call seconds after the last call
        """
        self.seconds = seconds
        self.trailing_call = trailing_call
        self.time_of_last_call = 0
        self.timer = None

    def __call__(self, fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                self.timer.cancel()
            except:
                pass

            def call_it():
                self.time_of_last_call = time.time()
                return fn(*args, **kwargs)

            seconds_since_last_call = time.time() - self.time_of_last_call
            if seconds_since_last_call >= self.seconds:
                return call_it()
            elif self.trailing_call:
                self.timer = threading.Timer(self.seconds - seconds_since_last_call, call_it)
                self.timer.start()

        return wrapper


def cache(ttl_seconds=100 * 365 * 24 * 3600, debug=False):
    """
    caching decorator with TTL. Keep in mind the cache is per-process.
    TODO: There is no process for cache invalidation now. Introduce memcached and use it instead.
    :param ttl_seconds: TTL in seconds before the cache entry expires
    :param debug: use True to log cache hits (with DEBUG level)
    """

    def decorate(f):
        _cache = {}

        @functools.wraps(f)
        def decorated(*args):
            cached = _cache.get(args)
            if not cached or cached["timestamp"] < datetime.datetime.now() - datetime.timedelta(
                seconds=ttl_seconds
            ):
                cached = {"timestamp": datetime.datetime.now(), "result": f(*args)}
                _cache[args] = cached
            elif debug:
                logger.debug(lambda: "@cache hit for %s" % str(args))
            return cached["result"]

        return decorated

    return decorate


class VarietyMetadata(GExiv2.Metadata):
    MULTIPLES = {
        "Iptc.Application2.Headline",
        "Iptc.Application2.Keywords",
        "Xmp.dc.creator",
        "Xmp.dc.subject",
    }

    NUMBERS = {
        "Xmp.variety.sfwRating",
        "Xmp.xmp.Rating",
        "Exif.Image.Rating",
        "Exif.Image.RatingPercent",
    }

    def __init__(self, path):
        super(VarietyMetadata, self).__init__(path=path)
        self.register_xmp_namespace("https://launchpad.net/variety/", "variety")

    def __getitem__(self, key):
        if self.has_tag(key):
            if key in self.MULTIPLES:
                return self.get_tag_multiple(key)
            elif key in self.NUMBERS:
                return self.get_tag_long(key)
            else:
                return self.get_tag_string(key)
        else:
            raise KeyError("%s: Unknown tag" % key)

    def __setitem__(self, key, value):
        if key in self.MULTIPLES:
            self.set_tag_multiple(key, value)
        elif key in self.NUMBERS:
            self.set_tag_long(key, value)
        else:
            self.set_tag_string(key, value)


class ModuleProfiler:
    # How deep in other modules' code we should profile
    MAX_NONTARGET_DEPTH = 1

    def __init__(self):
        """
        Initializes the module profiler.
        """
        self.target_paths = []

        # Track how far deep we are in functions outside our target packages
        # The intent is to only log the first call to outside methods without following them further
        self.nontarget_depths = {}

    def log_class(self, cls):
        """
        Adds the given class' module to the list of modules to be profiled.
        """
        modulename = cls.__module__
        if modulename not in sys.modules:
            logger.error(
                "ModuleProfiler: Could not add module %r (class %s) to the list of modules to trace - "
                "has it been imported entirely?",
                modulename,
                cls,
            )
            return

        module = sys.modules[modulename]

        self.log_module(module, request=cls)

    def log_module(self, module, request=None):
        """
        Adds the given module to the list of modules to be profiled.
        """
        self.log_path(module.__file__, request=request)

    def log_path(self, path, request=None):
        """
        Adds the given module path to the list of profile targets.
        """
        self.target_paths.append(path)

        logger.info(
            "ModuleProfiler: added path %s to list of profile targets (request=%s)", path, request
        )

    @functools.lru_cache(maxsize=2048)
    def is_target_path(self, path):
        """
        Returns whether the given path matches one of our modules to be profiled.
        """
        for target in self.target_paths:
            if os.path.isdir(target) and path.startswith(target + os.path.sep):
                return True
            elif path == target:
                return True
        return False

    def start(self):
        """
        Starts the module profiler for all future threads.
        """
        threading.setprofile(self.profiler)

    def stop(self):
        """
        Removes the module profiler globally and from future threads.
        """
        if sys.getprofile() != self.profiler:
            logger.warning(
                "ModuleProfiler: The currently enabled profile function was not ours - unbinding anyways"
            )
        threading.setprofile(None)
        sys.setprofile(None)

    def profiler(self, frame, event, arg):
        filename = frame.f_code.co_filename

        tid = threading.get_ident()

        if not self.is_target_path(filename):
            if tid not in self.nontarget_depths:
                # Pick up where the main thread left off
                self.nontarget_depths[tid] = self.nontarget_depths.get(
                    threading.main_thread().ident, 1
                )
            else:
                self.nontarget_depths[tid] += 1
        else:
            self.nontarget_depths[tid] = 0

        tname = threading.current_thread().name

        if event == "call":
            if self.nontarget_depths[tid] > self.MAX_NONTARGET_DEPTH:
                # Don't log past our max depth for packages that we're not tracking
                return
            else:
                # In order: function name, line number, filename
                s = "[%s] -> Entering function: %s\t(line %s in %s)" % (
                    tname,
                    frame.f_code.co_name,
                    frame.f_lineno,
                    filename,
                )
                if self.nontarget_depths[tid] == self.MAX_NONTARGET_DEPTH:
                    s += (
                        " - not tracing further because MAX_NONTARGET_DEPTH=%s"
                        % self.MAX_NONTARGET_DEPTH
                    )
                logger.debug(s)

        elif event == "return":
            if self.nontarget_depths[tid] > self.MAX_NONTARGET_DEPTH:
                return

            logger.debug(
                "[%s] -> Leaving function:  %s\t(line %s in %s)"
                % (tname, frame.f_code.co_name, frame.f_lineno, filename)
            )


class InternetDisabledError(Exception):
    pass


class Util:
    internet_enabled = True

    @staticmethod
    def sanitize_filename(filename):
        valid_chars = " ,.!-+@()_%s%s" % (string.ascii_letters, string.digits)
        return "".join(c if c in valid_chars else "_" for c in filename)

    @staticmethod
    def get_local_name(url, ensure_image=True):
        filename = url[url.rfind("/") + 1 :]
        index = filename.find("?")
        if index > 0:
            filename = filename[:index]
        index = filename.find("#")
        if index > 0:
            filename = filename[:index]

        filename = urllib.parse.unquote_plus(filename)

        filename = Util.sanitize_filename(filename)

        if len(filename) > 200:
            filename = filename[:190] + filename[-10:]

        if ensure_image and not Util.is_image(filename):
            filename += ".jpg"

        return filename

    @staticmethod
    def split(s, seps=(",", " ")):
        result = s.split()
        for sep in seps:
            result = [x.strip() for y in result for x in y.split(sep) if x.strip()]
        return result

    @staticmethod
    def makedirs(path):
        try:
            if not os.path.isdir(path):
                logger.info(lambda: "Creating folder %s" % path)
                os.makedirs(path)
        except OSError:
            logger.exception(lambda: "Could not makedirs for %s" % path)

    @staticmethod
    def is_image(filename, check_contents=False):
        if Util.is_animated_gif(filename):
            return False

        if not check_contents:
            return filename.lower().endswith(
                (".jpg", ".jpeg", ".gif", ".png", ".tiff", ".svg", ".bmp", ".avif")
            )
        else:
            format, image_width, image_height = GdkPixbuf.Pixbuf.get_file_info(filename)
            return bool(format)

    @staticmethod
    def is_animated_gif(filename):
        if not filename.lower().endswith(".gif"):
            return False

        gif = Image.open(filename)
        try:
            gif.seek(1)
        except EOFError:
            return False
        else:
            return True

    @staticmethod
    def list_files(
        files=(), folders=(), filter_func=(lambda f: True), max_files=10000, randomize=True
    ):
        class NextFolderException(Exception):
            pass

        count = 0
        for filepath in files:
            logger.debug(
                lambda: "checking file %s against filter_func %s" % (filepath, filter_func)
            )
            if filter_func(filepath) and os.access(filepath, os.R_OK):
                count += 1
                yield filepath

        folders = list(folders)
        if randomize:
            random.shuffle(folders)

        for folder in folders:
            folder_quota = max(20, int(max_files / len(folders)))
            if not os.path.isdir(folder):
                continue
            try:
                count_in_folder = 0
                for root, subfolders, files in os.walk(folder, followlinks=True):
                    subfolder_quota = max(10, int(folder_quota / (1 + len(subfolders))))
                    if randomize:
                        random.shuffle(files)
                        random.shuffle(subfolders)
                    for filename in files[:subfolder_quota]:
                        logger.debug(
                            lambda: "checking file %s against filter_func %s (root=%s)"
                            % (filename, filter_func, root)
                        )
                        path = os.path.join(root, filename)
                        if filter_func(path):
                            count += 1
                            if count > max_files:
                                logger.info(
                                    lambda: "More than %d files in the folders, stop listing"
                                    % max_files
                                )
                                return
                            yield path
                            count_in_folder += 1
                            if count_in_folder > folder_quota:
                                raise NextFolderException
            except NextFolderException:
                continue
            except Exception:
                logger.exception(lambda: "Could not walk folder " + folder)

    @staticmethod
    def start_force_exit_thread(delay):
        def force_exit():
            time.sleep(delay)
            print("Exiting takes too long. Calling os.kill.")
            os.kill(os.getpid(), 9)

        force_exit_thread = threading.Thread(target=force_exit)
        force_exit_thread.daemon = True
        force_exit_thread.start()

    @staticmethod
    def write_metadata(filename, info):
        try:
            m = VarietyMetadata(filename)
            for k, v in sorted(info.items()):
                if k == "author":
                    m["Xmp.variety." + k] = v
                    if not "Xmp.dc.creator" in m:
                        m["Xmp.dc.creator"] = [v]
                if k == "headline":
                    m["Iptc.Application2.Headline"] = [v]
                elif k == "description":
                    if v is not None:
                        m.set_comment(v)
                    else:
                        m.clear_comment()
                elif k == "keywords":
                    if not isinstance(v, (list, tuple)):
                        v = [v]
                    m["Iptc.Application2.Keywords"] = v
                    m["Xmp.dc.subject"] = v
                elif k == "sfwRating":
                    m["Xmp.variety." + k] = int(v)
                elif k == "extraData":
                    m["Xmp.variety." + k] = json.dumps(v, sort_keys=True)
                else:
                    m["Xmp.variety." + k] = v
            m.save_file()
            return True
        except Exception as ex:
            # could not write metadata inside file, use json instead
            logger.exception(
                lambda: "Could not write metadata directly in file, trying json metadata: "
                + filename
            )
            try:
                with open(filename + ".metadata.json", "w", encoding="utf8") as f:
                    f.write(json.dumps(info, indent=4, ensure_ascii=False, sort_keys=True))
                    return True
            except Exception as e:
                logger.exception(lambda: "Could not write metadata for file " + filename)
                return False

    @staticmethod
    def read_metadata(filename):
        try:
            m = VarietyMetadata(filename)

            info = {}
            for k in [
                "sourceName",
                "sourceLocation",
                "sourceURL",
                "sourceType",
                "imageURL",
                "author",
                "authorURL",
                "noOriginPage",
            ]:
                if "Xmp.variety." + k in m:
                    info[k] = m["Xmp.variety." + k]

            try:
                info["sfwRating"] = int(m["Xmp.variety.sfwRating"])
            except:
                pass

            try:
                info["author"] = m["Xmp.dc.creator"][0]
            except:
                pass

            try:
                info["headline"] = m["Iptc.Application2.Headline"][0]
            except:
                pass

            try:
                info["description"] = m.get_comment()
            except:
                pass

            try:
                info["extraData"] = json.loads(m["Xmp.variety.extraData"])
            except:
                pass

            try:
                info["keywords"] = m["Iptc.Application2.Keywords"]
            except:
                try:
                    info["keywords"] = m["Xmp.dc.subject"]
                except:
                    pass

            return info

        except Exception as e:
            # could not read metadata inside file, try reading json metadata instead
            try:
                with open(filename + ".metadata.json", encoding="utf8") as f:
                    return json.loads(f.read())

            except Exception:
                return None

    @staticmethod
    def set_rating(filename, rating):
        if rating is not None and (rating < -1 or rating > 5):
            raise ValueError("Rating should be between -1 and 5, or None")

        m = VarietyMetadata(filename)

        if rating is None:
            for key in ["Xmp.xmp.Rating", "Exif.Image.Rating", "Exif.Image.RatingPercent"]:
                if key in m:
                    del m[key]  # pylint: disable=unsupported-delete-operation
        else:
            m["Xmp.xmp.Rating"] = rating
            m["Exif.Image.Rating"] = max(0, rating)
            if rating >= 1:
                m["Exif.Image.RatingPercent"] = (rating - 1) * 25
            elif "Exif.Image.RatingPercent" in m:
                del m["Exif.Image.RatingPercent"]  # pylint: disable=unsupported-delete-operation

        m.save_file()

    @staticmethod
    def get_rating(filename):
        m = VarietyMetadata(filename)
        rating = None
        if "Xmp.xmp.Rating" in m:
            rating = m["Xmp.xmp.Rating"]
        elif "Exif.Image.Rating" in m:
            rating = m["Exif.Image.Rating"]
        elif "Exif.Image.RatingPercent" in m:
            rating = m["Exif.Image.RatingPercent"] // 25 + 1
        if rating is not None:
            rating = max(-1, min(5, rating))
        return rating

    @staticmethod
    def get_size(image):
        format, image_width, image_height = GdkPixbuf.Pixbuf.get_file_info(image)
        if not format:
            raise Exception("Not an image or unsupported image format")
        else:
            return image_width, image_height

    @staticmethod
    def get_primary_display_size(hidpi_scaled=True):
        display = Gdk.Display.get_default()
        monitor = display.get_primary_monitor()
        if not monitor:
            monitor = display.get_monitor(0)

        if monitor:
            geometry = monitor.get_geometry()
            scale = monitor.get_scale_factor() if hidpi_scaled else 1.0
            return int(geometry.width * scale), int(geometry.height * scale)
        else:
            return Util.get_multimonitor_display_size()

    @staticmethod
    def get_multimonitor_display_size():
        screen = Gdk.Screen.get_default()
        return screen.get_width(), screen.get_height()

    @staticmethod
    def find_unique_name(filename):
        index = filename.rfind(".")
        if index < 0:
            index = len(filename)
        before_extension = filename[:index]
        extension = filename[index:]
        i = 1
        f = filename
        while os.path.exists(f):
            f = before_extension + "_" + str(i) + extension
            i += 1
        return f

    @staticmethod
    def request(url, data=None, stream=False, method=None, timeout=5, headers=None):
        if not Util.internet_enabled:
            raise InternetDisabledError("Internet access in Variety is currently disabled")

        if url.startswith("//"):
            url = "http:" + url
        headers = headers or {}
        headers = {"User-Agent": USER_AGENT, "Cache-Control": "max-age=0", **headers}
        method = method if method else "POST" if data else "GET"
        try:
            r = requests.request(
                method=method,
                url=url,
                data=data,
                headers=headers,
                stream=stream,
                allow_redirects=True,
                timeout=timeout,
            )
            r.raise_for_status()
            return r
        except requests.exceptions.SSLError:
            logger.exception("SSL Error for url %s:" % url)
            raise

    @staticmethod
    def request_write_to(r, f):
        for chunk in r.iter_content(1024):
            f.write(chunk)

    @staticmethod
    def fetch(url, data=None, **request_kwargs):
        return Util.request(url, data, **request_kwargs).text

    @staticmethod
    def fetch_bytes(url, data=None, **request_kwargs):
        return Util.request(url, data, **request_kwargs).content

    @staticmethod
    def fetch_json(url, data=None, **request_kwargs):
        return Util.request(url, data, **request_kwargs).json()

    @staticmethod
    def html_soup(url, data=None, **request_kwargs):
        return bs4.BeautifulSoup(Util.fetch(url, data, **request_kwargs), "lxml")

    @staticmethod
    def xml_soup(url, data=None, **request_kwargs):
        return bs4.BeautifulSoup(Util.fetch(url, data, **request_kwargs), "xml")

    @staticmethod
    def unxor(text, key):
        ciphertext = base64.decodebytes(text)
        return "".join(chr(x ^ ord(y)) for (x, y) in zip(ciphertext, cycle(key)))

    @staticmethod
    def folderpath(folder):
        p = os.path.normpath(folder)
        if not p.endswith("/"):
            p += "/"
        return p

    @staticmethod
    def compute_trimmed_offsets(image_size, screen_size):
        """Computes what width or height of the wallpaper image will be trimmed on each side, as it is zoomed in to fill
        the whole screen. Returns a tuple (h, v, scale_ratio) in which h or v will be zero. The other one is the pixel
        width or height that will be trimmed on each one of the sides of the image (top and down or left and right)."""
        iw, ih = image_size
        screen_w, screen_h = screen_size
        screen_ratio = float(screen_w) / screen_h
        hoffset = voffset = 0
        if (
            screen_ratio > float(iw) / ih
        ):  # image is "taller" than the screen ratio - need to offset vertically
            scaledw = float(screen_w)
            scaledh = ih * scaledw / iw
            voffset = int((scaledh - float(scaledw) / screen_ratio) / 2)
        else:  # image is "wider" than the screen ratio - need to offset horizontally
            scaledh = float(screen_h)
            scaledw = iw * scaledh / ih
            hoffset = int((scaledw - float(scaledh) * screen_ratio) / 2)

        logger.info(
            lambda: "Trimmed offsets debug info: w:%d, h:%d, ratio:%f, iw:%d, ih:%d, scw:%d, sch:%d, ho:%d, vo:%d"
            % (screen_w, screen_h, screen_ratio, iw, ih, scaledw, scaledh, hoffset, voffset)
        )
        return hoffset, voffset

    @staticmethod
    def get_scaled_size(image):
        """Computes the size to which the image is scaled to fit the screen: original_size * scale_ratio = scaled_size"""
        iw, ih = Util.get_size(image)
        screen_w, screen_h = Util.get_primary_display_size()
        screen_ratio = float(screen_w) / screen_h
        if (
            screen_ratio > float(iw) / ih
        ):  # image is "taller" than the screen ratio - need to offset vertically
            return screen_w, int(round(ih * float(screen_w) / iw))
        else:  # image is "wider" than the screen ratio - need to offset horizontally
            return int(round(iw * float(screen_h) / ih)), screen_h

    @staticmethod
    def gtk_to_fcmatch_font(gtk_font_name):
        fd = Pango.FontDescription(gtk_font_name)
        family = fd.get_family()
        size = gtk_font_name[gtk_font_name.rindex(" ") :].strip()
        rest = gtk_font_name.replace(family, "").strip().replace(" ", ":")
        return family + ":" + rest, size

    @staticmethod
    def file_in(file, folder):
        return os.path.normpath(file).startswith(os.path.normpath(folder))

    @staticmethod
    def same_file_paths(f1, f2):
        return os.path.normpath(f1) == os.path.normpath(f2)

    @staticmethod
    def collapseuser(path):
        home = os.path.expanduser("~") + "/"
        return re.sub("^" + home, "~/", path)

    @staticmethod
    def compare_versions(v1, v2):
        from pkg_resources import parse_version

        pv1 = parse_version(v1)
        pv2 = parse_version(v2)

        if pv1 == pv2:
            return 0
        else:
            return -1 if pv1 < pv2 else 1

    @staticmethod
    def md5(s):
        if isinstance(s, str):
            s = s.encode("utf-8")
        return hashlib.md5(s).hexdigest()

    @staticmethod
    def md5file(file):
        with open(file, mode="rb") as f:
            return Util.md5(f.read())

    @staticmethod
    def random_hash():
        try:
            return codecs.encode(os.urandom(16), "hex_codec").decode("utf8")
        except Exception:
            return "".join(random.choice(string.hexdigits) for _ in range(32)).lower()

    @staticmethod
    def get_file_icon_name(path):
        try:
            f = Gio.File.new_for_path(os.path.normpath(os.path.expanduser(path)))
            query_info = f.query_info("standard::icon", Gio.FileQueryInfoFlags.NONE, None)
            return query_info.get_attribute_object("standard::icon").get_names()[0]
        except Exception:
            logger.exception(lambda: "Exception while obtaining folder icon for %s:" % path)
            return "folder"

    @staticmethod
    def is_home_encrypted():
        return os.path.isdir(os.path.expanduser("~").replace("/home/", "/home/.ecryptfs/"))

    @staticmethod
    def get_xdg_pictures_folder():
        try:
            pics_folder = GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES)
            if not pics_folder:
                raise Exception("Could not get path to Pictures folder. Defaulting to ~/Pictures.")
            return pics_folder
        except:
            logger.exception(
                lambda: "Could not get path to Pictures folder. Defaulting to ~/Pictures."
            )
            return os.path.expanduser("~/Pictures")

    @staticmethod
    def superuser_exec(*command_args):
        logger.warning(lambda: "Executing as superuser: %s" % command_args)
        subprocess.check_call(["pkexec"] + list(command_args))

    @staticmethod
    def safe_map(f, l):
        for element in l:
            try:
                yield f(element)
            except Exception:
                continue

    @staticmethod
    def get_thumbnail_data(image, width, height):
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(image, width, height)
        return pixbuf.save_to_bufferv("jpeg", [], [])[1]

    @staticmethod
    def is_alive_and_image(url):
        try:
            r = Util.request(url, method="head")
            return r.headers.get("content-type", "").startswith("image/")
        except Exception:
            return False

    @staticmethod
    def is_dead_or_not_image(url):
        if not url:
            return True

        try:
            host = urllib.parse.urlparse(url).netloc
            if host.startswith("interfacelift.com"):
                return False

            if "wallbase.cc" in host or "ns223506.ovh.net" in host:
                return True
        except:
            return True

        try:
            r = Util.request(url, method="head")
            return not r.headers.get("content-type", "").startswith("image/")
        except requests.exceptions.RequestException:
            return True
        except:
            return False

    # makes the Gtk thread execute the given callback.
    @staticmethod
    def add_mainloop_task(callback, *args):
        def cb(args):
            args[0](*args[1:])
            return False

        args = [callback] + list(args)
        Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT, cb, args)

    @staticmethod
    def is_unity():
        return os.getenv("XDG_CURRENT_DESKTOP", "").lower() == "unity"

    @staticmethod
    def start_daemon(target):
        daemon_thread = threading.Thread(target=target)
        daemon_thread.daemon = True
        daemon_thread.start()
        return daemon_thread

    @staticmethod
    def check_variety_slideshow_present():
        return bool(shutil.which("variety-slideshow"))

    @staticmethod
    def convert_to_filename(url):
        url = re.sub(r"http://", "", url)
        url = re.sub(r"https://", "", url)
        valid_chars = "_%s%s" % (string.ascii_letters, string.digits)
        return "".join(c if c in valid_chars else "_" for c in url)

    @staticmethod
    def safe_unlink(filepath):
        try:
            os.unlink(filepath)
        except Exception:
            logger.exception(lambda: "Could not delete {}, ignoring".format(filepath))

    @staticmethod
    def copy_with_replace(from_path, to_path, search_replace_map):
        with open(from_path, "r") as file:
            data = file.read()
        for search, replace in search_replace_map.items():
            data = data.replace(search, replace)
        with open(to_path + ".partial", "w") as file:
            file.write(data)
            file.flush()
        os.rename(to_path + ".partial", to_path)

    @staticmethod
    def get_exec_path():
        return os.path.abspath(sys.argv[0])

    @staticmethod
    def get_folder_size(start_path):
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(start_path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if not os.path.islink(fp):
                    total_size += os.path.getsize(fp)
        return total_size


def on_gtk(f):
    @functools.wraps(f)
    def wrapped(*args):
        Util.add_mainloop_task(f, *args)

    return wrapped


def safe_print(text, ascii_text=None, file=sys.stdout):
    """
    Python's print throws UnicodeEncodeError if the terminal encoding is borked. This version tries print, then logging, then printing the ascii text when one is present.
    If does not throw exceptions even if it fails.
    :param text: Text to print, str or unicode, possibly with non-ascii symbols in it
    :param ascii_text: optional. Original untranslated ascii version of the text when present.
    """
    try:
        print(text, file=file)
    except:  # UnicodeEncodeError can happen here if the terminal is strangely configured, but we are playing safe and catching everything
        try:
            logging.getLogger("variety").error(
                "Error printing non-ascii text, terminal encoding is %s" % sys.stdout.encoding
            )
            if ascii_text:
                try:
                    print(ascii_text)
                    return
                except:
                    pass
            logging.getLogger("variety").warning(text)
        except:
            pass



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import random

from variety.plugins.builtin.downloaders.MediaRSSDownloader import MediaRSSDownloader
from variety.plugins.downloaders.ConfigurableImageSource import ConfigurableImageSource
from variety.Util import _

logger = logging.getLogger("variety")

random.seed()


class MediaRSSSource(ConfigurableImageSource):
    @classmethod
    def get_info(cls):
        return {
            "name": "MediaRSSSource",
            "description": _("Configurable source for fetching images from MediaRSS feeds"),
            "author": "Peter Levi",
            "version": "0.1",
        }

    def get_source_name(self):
        return "MediaRSS"

    def get_source_type(self):
        return "mediarss"

    def get_ui_instruction(self):
        return _(
            "Please paste the URL of the Media RSS feed below. Please note that only Media RSS "
            "feeds are supported, not arbitrary RSS feeds. Media RSS feeds contain media:content "
            "tags linking directly to the actual image content. "
            "Some examples of sites that provide Media RSS feeds are: "
            "<a href='https://picasaweb.google.com/'>Picasa</a>, "
            "<a href='http://www.deviantart.com'>deviantART</a>, "
            "<a href='http://www.smugmug.com/browse/'>SmugMug</a>, "
            "<a href='http://www.flickr.com'>Flickr</a>, "
            "<a href='http://interfacelift.com'>InterfaceLIFT</a>."
        )

    def get_ui_short_instruction(self):
        return _("Paste the URL of the Media RSS feed here: ")

    def get_ui_short_description(self):
        return _("Fetch images from a MediaRSS feed")

    def validate(self, url):
        if not url.startswith("http://") and not url.startswith("https://"):
            url = "https://" + url
        valid = MediaRSSDownloader.validate(url)
        error = _(
            "This does not seem to be a valid Media RSS feed URL or there is no content there."
        )
        return url, None if valid else error

    def create_downloader(self, config):
        return MediaRSSDownloader(self, config)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012-2018, Peter Levi <peterlevi@peterlevi.com>
# Copyright (c) 2017-2018, James Lu <james@overdrivenetworks.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import dbus, dbus.service, dbus.glib
import logging
import os
import signal
import sys

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # pylint: disable=E0611


class SafeLogger(logging.Logger):
    """
    Fixes UnicodeDecodeErrors errors in logging calls:
    Accepts lambda as well string messages. Catches errors when evaluating the passed lambda.
    """

    def makeRecord(self, name, level, fn, lno, msg, *args, **kwargs):
        try:
            new_msg = msg if isinstance(msg, str) else msg()
        except:
            locale_info = "Unknown"
            try:
                locale_info = "Terminal encoding=%s, LANG=%s, LANGUAGE=%s" % (
                    sys.stdout.encoding,
                    os.getenv("LANG"),
                    os.getenv("LANGUAGE"),
                )
                logging.getLogger("variety").exception(
                    "Errors while logging. Locale info: %s" % locale_info
                )
                # TODO gather and log more info here
            except:
                pass
            new_msg = "Errors while logging. Locale info: %s" % locale_info

        return super().makeRecord(name, level, fn, lno, new_msg, *args, **kwargs)


logging.setLoggerClass(SafeLogger)


# these must be after the setLoggerClass call, as they obtain the variety logger
from variety import VarietyWindow, ThumbsManager, ThumbsWindow
from variety.profile import set_profile_path, get_profile_path, is_default_profile, get_profile_id
from variety.Util import Util, _, ModuleProfiler, safe_print


# # Change default encoding from ascii to UTF8 - works OK on Linux and prevents various UnicodeEncodeErrors/UnicodeDecodeErrors
# Still, generally considerd bad practice, may cause some deep hidden errors, as various Python stuff depends on it
# reload(sys)
# sys.setdefaultencoding('UTF8')


def _get_dbus_key():
    """
    DBus key for Variety.
    Variety uses a different key per profile, so several instances can run simultaneously if
    running with different profiles.
    Command any instance from the terminal by passing explicitly the same --profile options as it
    was started with.
    :return: the dbus key
    """
    if is_default_profile():
        return "com.peterlevi.Variety"
    else:
        return "com.peterlevi.Variety_{}".format(get_profile_id())


DBUS_PATH = "/com/peterlevi/Variety"


class VarietyService(dbus.service.Object):
    def __init__(self, variety_window):
        self.variety_window = variety_window
        bus_name = dbus.service.BusName(_get_dbus_key(), bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, bus_name, DBUS_PATH)

    @dbus.service.method(dbus_interface=_get_dbus_key(), in_signature="as", out_signature="s")
    def process_command(self, arguments):
        result = self.variety_window.process_command(arguments, initial_run=False)
        return "" if result is None else result


VARIETY_WINDOW = None

terminate = False


def _sigint_handler(*args):
    global terminate
    terminate = True


def _check_quit():
    global terminate
    if not terminate:
        GObject.timeout_add(1000, _check_quit)
        return

    logging.getLogger("variety").info("Terminating signal received, quitting...")
    safe_print(
        _("Terminating signal received, quitting..."),
        "Terminating signal received, quitting...",
        file=sys.stderr,
    )

    global VARIETY_WINDOW
    if VARIETY_WINDOW:
        VARIETY_WINDOW.on_quit()
    Util.start_force_exit_thread(10)


def _set_up_logging(verbose):
    # add a handler to prevent basicConfig
    root = logging.getLogger()
    null_handler = logging.NullHandler()
    root.addHandler(null_handler)

    formatter = logging.Formatter("%(levelname)s: %(asctime)s: %(funcName)s() '%(message)s'")

    logger = logging.getLogger("variety")
    logger_sh = logging.StreamHandler()
    logger_sh.setFormatter(formatter)
    logger.addHandler(logger_sh)

    try:
        logger_file = logging.FileHandler(os.path.join(get_profile_path(), "variety.log"), "w")
        logger_file.setFormatter(formatter)
        logger.addHandler(logger_file)
    except Exception:
        logger.exception("Could not create file logger")

    lib_logger = logging.getLogger("variety_lib")
    lib_logger_sh = logging.StreamHandler()
    lib_logger_sh.setFormatter(formatter)
    lib_logger.addHandler(lib_logger_sh)

    logger.setLevel(logging.INFO)
    # Set the logging level to show debug messages.
    if verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif not verbose:
        # If we're not in verbose mode, only log these messages to file. This prevents
        # flooding syslog and/or ~/.xsession-errors depending on how variety was started:
        # (https://bugs.launchpad.net/variety/+bug/1685003)
        # XXX: We should /really/ make the internal debug logging use logging.debug,
        # this is really just a bandaid patch.
        logger_sh.setLevel(logging.WARNING)

    if verbose >= 3:
        lib_logger.setLevel(logging.DEBUG)


def main():
    # Ctrl-C
    signal.signal(signal.SIGINT, _sigint_handler)
    signal.signal(signal.SIGTERM, _sigint_handler)
    signal.signal(signal.SIGQUIT, _sigint_handler)

    arguments = sys.argv[1:]

    # validate arguments
    from variety import VarietyOptionParser

    options, args = VarietyOptionParser.parse_options(arguments)
    set_profile_path(options.profile)
    Util.makedirs(get_profile_path())

    # ensure singleton per profile
    bus = dbus.SessionBus()
    dbus_key = _get_dbus_key()
    if bus.request_name(dbus_key) != dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER:
        if not arguments or (options.profile and len(arguments) <= 2):
            arguments = ["--preferences"]
        safe_print(
            _("Variety is already running. Sending the command to the running instance."),
            "Variety is already running. Sending the command to the running instance.",
            file=sys.stderr,
        )
        method = bus.get_object(dbus_key, DBUS_PATH).get_dbus_method("process_command")
        result = method(arguments)
        if result:
            safe_print(result)
        return

    # set up logging
    # set_up_logging must be called after the DBus checks, only by one running instance,
    # or the log file can be corrupted
    _set_up_logging(options.verbose)
    logging.getLogger("variety").info(lambda: "Using profile folder {}".format(get_profile_path()))

    if options.verbose >= 3:
        profiler = ModuleProfiler()
        if options.verbose >= 5:
            # The main variety package
            pkgname = os.path.dirname(__file__)
            profiler.log_path(pkgname)

            if options.verbose >= 6:
                # Track variety_lib
                profiler.log_path(pkgname + "_lib")
        else:
            # Cherry-picked log items carried over from variety 0.6.x
            profiler.log_class(VarietyWindow.VarietyWindow)

            if options.verbose >= 4:
                profiler.log_class(ThumbsManager.ThumbsManager)
                profiler.log_class(ThumbsWindow.ThumbsWindow)

        profiler.start()

    # Run the application.
    window = VarietyWindow.VarietyWindow()
    global VARIETY_WINDOW
    VARIETY_WINDOW = window
    service = VarietyService(window)

    bus.call_on_disconnection(window.on_quit)

    window.start(arguments)
    GObject.timeout_add(2000, _check_quit)
    Gtk.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random
import urllib.parse
import xml.etree.ElementTree as ET

from variety.plugins.downloaders.DefaultDownloader import DefaultDownloader
from variety.Util import Util

logger = logging.getLogger("variety")

random.seed()

MEDIA_NS = "{http://search.yahoo.com/mrss/}"
VARIETY_NS = "{http://vrty.org/}"


class MediaRSSDownloader(DefaultDownloader):
    def __init__(self, source, url):
        DefaultDownloader.__init__(self, source=source, config=url)

    @staticmethod
    def fetch(url):
        content = Util.fetch_bytes(url)
        return ET.fromstring(content)

    @staticmethod
    def is_valid_content(x):
        return (
            x is not None
            and "url" in x.attrib
            and (
                Util.is_image(x.attrib["url"])
                or ("medium" in x.attrib and x.attrib["medium"].lower() == "image")
                or ("type" in x.attrib and x.attrib["type"].lower().startswith("image/"))
            )
        )

    @staticmethod
    def validate(url):
        logger.info(lambda: "Validating MediaRSS url " + url)
        try:
            if not url.startswith("http://") and not url.startswith("https://"):
                url = "https://" + url

            s = MediaRSSDownloader.fetch(url)
            walls = [
                x.attrib["url"]
                for x in s.findall(".//{0}content".format(MEDIA_NS))
                if MediaRSSDownloader.is_valid_content(x)
            ]
            return len(walls) > 0
        except Exception:
            logger.exception(lambda: "Error while validating URL, probably not a MediaRSS feed")
            return False

    def download_queue_item(self, queue_item):
        origin_url, image_url, source_type, source_location, source_name, extra_metadata = (
            queue_item
        )
        parse = urllib.parse.urlparse(origin_url)
        host = parse.netloc if hasattr(parse, "netloc") else "origin"
        return self.save_locally(
            origin_url,
            image_url,
            source_type or "mediarss",
            source_location,
            source_name or host,
            extra_metadata=extra_metadata,
        )

    def fill_queue(self):
        queue = []
        logger.info(lambda: "MediaRSS URL: " + self.config)
        s = self.fetch(self.config)

        for item in s.findall(".//item"):
            try:
                origin_url = item.find("link").text
                group = item.find("{0}group".format(MEDIA_NS))
                content = None
                width = -1
                if group is not None:
                    # find the largest image in the group
                    for c in group.findall("{0}content".format(MEDIA_NS)):
                        try:
                            if MediaRSSDownloader.is_valid_content(c):
                                if content is None:
                                    content = (
                                        c
                                    )  # use the first one, in case we don't find any width info
                                if "width" in c.attrib and int(c.attrib["width"]) > width:
                                    content = c
                                    width = int(c.attrib["width"])
                        except Exception:
                            pass
                else:
                    content = item.find("{0}content".format(MEDIA_NS))

                if not MediaRSSDownloader.is_valid_content(content):
                    continue

                source_name = None
                source_location = None
                source_type = None
                variety_source = item.find("{0}source".format(VARIETY_NS))
                if variety_source is not None:
                    source_name = variety_source.attrib.get("name", None)
                    source_location = variety_source.attrib.get("location", None)
                    source_type = variety_source.attrib.get("type", None)

                extra_metadata = {}

                try:
                    extra_metadata["headline"] = item.find("{0}title".format(MEDIA_NS)).text
                except:
                    try:
                        extra_metadata["headline"] = item.find("title").text
                    except:
                        pass

                try:
                    extra_metadata["description"] = item.find(
                        "{0}description".format(MEDIA_NS)
                    ).text
                except:
                    pass

                try:
                    author = item.find("{0}author".format(VARIETY_NS))
                    if author is not None:
                        extra_metadata["author"] = author.attrib.get("name", None)
                        extra_metadata["authorURL"] = author.attrib.get("url", None)
                    else:
                        extra_metadata["author"] = item.find("{0}credit".format(MEDIA_NS)).text
                except:
                    pass

                try:
                    sfw = item.find("{0}sfw_info".format(VARIETY_NS))
                    if sfw is not None:
                        rating = int(sfw.attrib.get("rating", None))
                        extra_metadata["sfwRating"] = rating

                        if self.is_safe_mode_enabled() and rating < 100:
                            logger.info(
                                lambda: "Skipping non-safe download from VRTY MediaRss feed. "
                                "Is the source %s suitable for Safe mode?" % self.config
                            )
                            continue
                except:
                    pass

                try:
                    extra_metadata["keywords"] = [
                        k.strip() for k in item.find("{0}keywords".format(MEDIA_NS)).text.split(",")
                    ]
                except:
                    pass

                self.process_content(
                    queue,
                    origin_url,
                    content,
                    source_type,
                    source_location,
                    source_name,
                    extra_metadata,
                )
            except Exception:
                logger.exception(lambda: "Could not process an item in the Media RSS feed")

        random.shuffle(queue)
        return queue

    def process_content(
        self,
        queue,
        origin_url,
        content,
        source_type=None,
        source_location=None,
        source_name=None,
        extra_metadata={},
    ):
        try:
            logger.debug(lambda: "Checking origin_url " + origin_url)

            if self.is_in_banned(origin_url):
                logger.debug(lambda: "In banned, skipping")
                return

            image_file_url = content.attrib["url"]

            if self.is_in_downloaded(image_file_url):
                logger.debug(lambda: "Already in downloaded")
                return

            if self.is_in_favorites(image_file_url):
                logger.debug(lambda: "Already in favorites")
                return

            width = None
            height = None
            try:
                width = int(content.attrib["width"])
                height = int(content.attrib["height"])
            except Exception:
                pass

            if self.is_size_inadequate(width, height):
                logger.debug(lambda: "Small or non-landscape size/resolution")
                return

            logger.debug(
                lambda: "Appending to queue %s, %s, %s, %s, %s"
                % (origin_url, image_file_url, source_type, source_location, source_name)
            )
            queue.append(
                (
                    origin_url,
                    image_file_url,
                    source_type,
                    source_location,
                    source_name,
                    extra_metadata,
                )
            )
        except Exception:
            logger.exception(lambda: "Error parsing single MediaRSS image info:")



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random
import urllib.parse

from variety.plugins.downloaders.DefaultDownloader import DefaultDownloader
from variety.plugins.downloaders.ImageSource import ImageSource, Throttling
from variety.Util import Util

logger = logging.getLogger("variety")

random.seed()

API_KEY = "d9c4a1fd59926d27d091b69651c37a34"
HASH = b"VAxWBwAJUlxWCQBQVQJWBVVUClBbDg1SDAVTC1RRB1E=\n"


class FlickrDownloader(ImageSource, DefaultDownloader):
    def __init__(self, parent, location):
        ImageSource.__init__(self)
        DefaultDownloader.__init__(self, source=self, config=location)
        self.set_variety(parent)
        self.parse_location()

    @classmethod
    def get_info(cls):
        raise Exception("Not yet implemented as a plugin")

    def get_description(self):
        return _("Images from Flickr")

    def get_source_name(self):
        return "Flickr"

    def get_source_type(self):
        return "flickr"

    def get_default_throttling(self):
        return Throttling(max_downloads_per_hour=120, max_queue_fills_per_hour=20)

    def parse_location(self):
        s = self.config.split(";")
        self.params = {}
        for x in s:
            if len(x) and x.find(":") > 0:
                k, v = x.split(":")
                if k.lower() in ["text", "tags", "user_id", "group_id"]:
                    self.params[k.lower()] = v.replace(" ", "+")

        # slight validation:
        for k in ["text", "tags", "user_id", "group_id"]:
            if k in self.params and len(self.params[k]) > 0:
                return
        raise Exception("Missing at least one of text, tags, user_id and group_id")

    @staticmethod
    def fetch(call):
        logger.info(lambda: "Making flickr API call: " + call)
        return Util.fetch_json(call)

    @staticmethod
    def obtain_userid(url):
        try:
            logger.info(lambda: "Fetching flickr user_id from URL: " + url)

            call = (
                "https://api.flickr.com/services/rest/?method=flickr.urls.lookupUser&api_key=%s&url=%s&format=json&nojsoncallback=1"
                % (Util.unxor(HASH, API_KEY), urllib.parse.quote_plus(url))
            )

            resp = FlickrDownloader.fetch(call)

            if resp["stat"] == "ok":
                logger.info(lambda: "Found " + resp["user"]["id"])
                return True, "ok", resp["user"]["id"]
            else:
                logger.info(lambda: "Oops " + resp["message"])
                return False, resp["message"], None
        except Exception as e:
            logger.exception(lambda: "Exception while checking Flickr user")
            return (False, "Exception while checking user. Please run with -v and check log.", None)

    @staticmethod
    def obtain_groupid(url):
        try:
            logger.info(lambda: "Fetching flickr group_id from URL: " + url)

            call = (
                "https://api.flickr.com/services/rest/?method=flickr.urls.lookupGroup&api_key=%s&url=%s&format=json&nojsoncallback=1"
                % (Util.unxor(HASH, API_KEY), urllib.parse.quote_plus(url))
            )

            resp = FlickrDownloader.fetch(call)

            if resp["stat"] == "ok":
                logger.info(lambda: "Found " + resp["group"]["id"])
                return True, "ok", resp["group"]["id"]
            else:
                logger.info(lambda: "Oops " + resp["message"])
                return False, resp["message"], None
        except Exception as e:
            logger.exception(lambda: "Exception while checking Flickr group")
            return (
                False,
                "Exception while checking group. Please run with -v and check log.",
                None,
            )

    @staticmethod
    def count_search_results(search):
        try:
            dl = FlickrDownloader(None, search)
            return dl.count_results()
        except Exception:
            logger.exception(lambda: "Exception while counting Flickr results")
            return 0

    def count_results(self):
        call = (
            "https://api.flickr.com/services/rest/?method=flickr.photos.search"
            "&api_key=%s&per_page=20&tag_mode=all&format=json&nojsoncallback=1"
            % Util.unxor(HASH, API_KEY)
        )

        for k, v in self.params.items():
            call = call + "&" + k + "=" + v

        resp = FlickrDownloader.fetch(call)
        if resp["stat"] != "ok":
            raise Exception("Flickr returned error message: " + resp["message"])

        return int(resp["photos"]["total"])

    def fill_queue(self):
        queue = []

        call = (
            "https://api.flickr.com/services/rest/?method=flickr.photos.search"
            "&api_key=%s&per_page=500&tag_mode=all&format=json&nojsoncallback=1"
            % Util.unxor(HASH, API_KEY)
        )

        for k, v in self.params.items():
            call = call + "&" + k + "=" + v

        resp = FlickrDownloader.fetch(call)
        if resp["stat"] != "ok":
            raise Exception("Flickr returned error message: " + resp["message"])

        pages = int(resp["photos"]["pages"])
        if pages < 1:
            return

        page = random.randint(1, pages)
        logger.info(lambda: "%d pages in the search results, using page %d" % (pages, page))

        call = (
            call
            + "&extras=owner_name,description,tags,o_dims,url_o,url_k,url_h,url_l&page="
            + str(page)
        )
        resp = FlickrDownloader.fetch(call)
        if resp["stat"] != "ok":
            raise Exception("Flickr returned error message: " + resp["message"])

        used = set(x[0] for x in queue)
        size_suffixes = ["o", "k", "h", "l"]
        for s in size_suffixes:
            self.process_photos_in_response(queue, resp, s, used)
            if len(queue) > 20:
                break

        random.shuffle(queue)
        if len(queue) >= 20:
            queue = queue[: len(queue) // 2]
            # only use randomly half the images from the page -
            # if we ever hit that same page again, we'll still have what to download

        return queue

    def process_photos_in_response(self, queue, resp, size_suffix, used):
        logger.info(
            lambda: "Queue size is %d, populating with images for size suffix %s"
            % (len(queue), size_suffix)
        )
        for ph in resp["photos"]["photo"]:
            try:
                photo_url = "https://www.flickr.com/photos/%s/%s" % (ph["owner"], ph["id"])
                logger.debug(lambda: "Checking photo_url " + photo_url)

                if self.is_in_banned(photo_url):
                    logger.debug(lambda: "In banned, skipping")
                    continue
                if photo_url in used:
                    logger.debug(lambda: "Already added or checked, skipping")
                    continue

                if "url_" + size_suffix in ph:
                    width = int(ph["width_" + size_suffix])
                    height = int(ph["height_" + size_suffix])
                    image_file_url = ph["url_" + size_suffix]
                    logger.debug(lambda: "Image url: " + image_file_url)
                else:
                    logger.debug(lambda: "Missing size " + size_suffix)
                    continue

                # add to used now - if one of the checks below fails, we don't want the lower resolutions either
                used.add(photo_url)

                if self.is_in_downloaded(image_file_url):
                    logger.debug(lambda: "Already in downloaded")
                    continue

                if self.is_in_favorites(image_file_url):
                    logger.debug(lambda: "Already in favorites")
                    continue

                if self.is_size_inadequate(width, height):
                    logger.debug(lambda: "Small or non-landscape size/resolution")
                    continue

                try:
                    extra_metadata = {
                        "author": ph["ownername"],
                        "authorURL": "https://www.flickr.com/photos/%s" % ph["owner"],
                        "headline": ph["title"],
                        "keywords": ph["tags"].split(" ")[
                            :200
                        ],  # Flickr metadata can be excessive and hit Exif limits
                        "description": ph["description"]["_content"][:10000],
                    }
                except:
                    extra_metadata = {}

                logger.debug(lambda: "Appending to queue %s, %s" % (photo_url, image_file_url))
                queue.append((photo_url, image_file_url, extra_metadata))
            except Exception:
                logger.exception(lambda: "Error parsing single flickr photo info:")

    @staticmethod
    def get_photo_id(origin_url):
        if origin_url[-1] == "/":
            origin_url = origin_url[:-1]
        return origin_url.split("/")[-1]

    @staticmethod
    def get_image_url(origin_url):
        photo_id = FlickrDownloader.get_photo_id(origin_url)
        call = (
            "https://api.flickr.com/services/rest/?method=flickr.photos.getSizes&api_key=%s&photo_id=%s&format=json&nojsoncallback=1"
            % (Util.unxor(HASH, API_KEY), photo_id)
        )
        resp = Util.fetch_json(call)
        s = max(resp["sizes"]["size"], key=lambda size: int(size["width"]))
        return s["source"]

    @staticmethod
    def get_extra_metadata(origin_url):
        photo_id = FlickrDownloader.get_photo_id(origin_url)
        call = (
            "https://api.flickr.com/services/rest/?method=flickr.photos.getInfo&api_key=%s&photo_id=%s&format=json&nojsoncallback=1"
            % (Util.unxor(HASH, API_KEY), photo_id)
        )
        resp = Util.fetch_json(call)
        ph = resp["photo"]
        extra_meta = {
            "headline": ph["title"]["_content"],
            "description": ph["description"]["_content"],
            "author": ph["owner"]["realname"],
            "authorURL": "https://www.flickr.com/photos/%s" % ph["owner"]["nsid"],
            "keywords": [x["_content"] for x in ph["tags"]["tag"]],
        }
        return extra_meta



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

from gi.repository import Gtk  # pylint: disable=E0611

from variety.profile import get_profile_wm_class

from .helpers import get_builder


class AboutDialog(Gtk.AboutDialog):
    __gtype_name__ = "AboutDialog"

    def __new__(cls):
        """Special static method that's automatically called by Python when 
        constructing a new instance of this class.
        
        Returns a fully instantiated AboutDialog object.
        """
        builder = get_builder("AboutVarietyDialog")
        new_object = builder.get_object("about_variety_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initalizing should be called after parsing the ui definition
        and creating a AboutDialog object with it in order
        to finish initializing the start of the new AboutVarietyDialog
        instance.
        
        Put your initialization code in here and leave __init__ undefined.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)
        self.set_wmclass(get_profile_wm_class(), get_profile_wm_class())



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""Code to add AppIndicator."""

import logging
import os

from gi.repository import Gtk  # pylint: disable=E0611

from variety.Util import Util, _
from variety_lib import varietyconfig

THEME_ICON_NAME = "variety-indicator"
THEME_ICON_NAME_DARK = "variety-indicator-dark"
THEME_ICON_NAME_NUM = "variety-indicator-num{}"

try:
    import gi

    try:
        gi.require_version("AyatanaAppIndicator3", "0.1")
        from gi.repository import AyatanaAppIndicator3 as AppIndicator3  # pylint: disable=E0611

        _indicator_backend = "AyatanaAppIndicator3"  # Just a dummy value we use for logging
    except (ValueError, ImportError):
        gi.require_version("AppIndicator3", "0.1")
        from gi.repository import AppIndicator3  # pylint: disable=E0611

        _indicator_backend = "AppIndicator3"
    use_appindicator = True
except (ValueError, ImportError):
    _indicator_backend = "fallback tray"
    use_appindicator = False


logger = logging.getLogger("variety")


class Indicator:
    def __init__(self, window):
        self.parent = window
        self.create_menu(window)
        self.create_indicator(window)

    def create_menu(self, window):
        self.menu = Gtk.Menu()

        if not Util.is_unity():
            self.next_main = Gtk.MenuItem(_("_Next"))
            self.next_main.set_use_underline(True)
            self.next_main.connect("activate", window.next_wallpaper)
            self.menu.append(self.next_main)

            self.prev_main = Gtk.MenuItem(_("_Previous"))
            self.prev_main.set_use_underline(True)
            self.prev_main.connect("activate", window.prev_wallpaper)
            self.menu.append(self.prev_main)

            self.menu.append(Gtk.SeparatorMenuItem.new())

        self.file_label = Gtk.MenuItem(_("Current desktop wallpaper"))
        self.file_label.connect("activate", window.open_file)
        self.menu.append(self.file_label)

        self.show_origin = Gtk.MenuItem(_("Show origin"))
        self.show_origin.connect("activate", window.on_show_origin)
        self.show_origin.set_sensitive(False)
        self.menu.append(self.show_origin)

        self.show_author = Gtk.MenuItem("Show author")
        self.show_author.connect("activate", window.on_show_author)
        self.show_author.set_sensitive(False)
        self.show_author.set_visible(False)
        self.menu.append(self.show_author)

        self.copy_to_favorites = Gtk.MenuItem(_("Copy to _Favorites"))
        self.copy_to_favorites.set_use_underline(True)
        self.copy_to_favorites.connect("activate", window.copy_to_favorites)
        self.menu.append(self.copy_to_favorites)

        self.move_to_favorites = Gtk.MenuItem(_("Move to Favorites"))
        self.move_to_favorites.set_use_underline(True)
        self.move_to_favorites.connect("activate", window.move_to_favorites)
        self.move_to_favorites.set_visible(False)
        self.menu.append(self.move_to_favorites)

        self.trash = Gtk.MenuItem(_("Delete to _Trash"))
        self.trash.set_use_underline(True)
        self.trash.connect("activate", window.move_to_trash)
        self.menu.append(self.trash)

        self.menu.append(Gtk.SeparatorMenuItem.new())

        self.image_menu = Gtk.Menu()

        self.next = Gtk.MenuItem(_("_Next"))
        self.next.set_use_underline(True)
        self.next.connect("activate", window.next_image_same_quote)
        self.image_menu.append(self.next)

        self.prev = Gtk.MenuItem(_("_Previous"))
        self.prev.set_use_underline(True)
        self.prev.connect("activate", window.prev_wallpaper)
        self.image_menu.append(self.prev)

        self.fast_forward = Gtk.MenuItem(_("_Next, skipping forward history"))
        self.fast_forward.set_use_underline(True)

        def _fast_forward(widget):
            window.next_wallpaper(widget, bypass_history=True)

        self.fast_forward.connect("activate", _fast_forward)
        self.image_menu.append(self.fast_forward)

        self.image_menu.append(Gtk.SeparatorMenuItem.new())
        self.scroll_tip = Gtk.MenuItem(_("Tip: Scroll wheel over icon\nfor Next and Previous"))
        self.scroll_tip.set_sensitive(False)
        self.image_menu.append(self.scroll_tip)

        self.image_menu.append(Gtk.SeparatorMenuItem.new())

        self.pause_resume = Gtk.MenuItem(_("Pause on current"))
        self.pause_resume.connect("activate", window.on_pause_resume)
        self.image_menu.append(self.pause_resume)

        self.image_item = Gtk.MenuItem(_("_Image"))
        self.image_item.set_use_underline(True)
        self.image_item.set_submenu(self.image_menu)
        self.menu.append(self.image_item)

        self.image_menu.append(Gtk.SeparatorMenuItem.new())

        self.focus = Gtk.MenuItem(_("Where is it from?"))
        self.focus.connect("activate", window.focus_in_preferences)
        self.image_menu.append(self.focus)

        self.no_effects = Gtk.CheckMenuItem(_("Show without effects"))
        self.no_effects.set_active(False)
        self.no_effects.set_use_underline(True)

        def _toggle_no_effects(widget=None):
            window.toggle_no_effects(self.no_effects.get_active())

        self.no_effects_handler_id = self.no_effects.connect("toggled", _toggle_no_effects)
        self.image_menu.append(self.no_effects)

        self.google_image = Gtk.MenuItem(_("Google Image Search"))
        self.google_image.connect("activate", window.google_image_search)
        self.image_menu.append(self.google_image)

        self.rating_separator = Gtk.SeparatorMenuItem.new()
        self.image_menu.append(self.rating_separator)

        self.rating = Gtk.MenuItem(_("Set EXIF Rating"))
        self.image_menu.append(self.rating)

        # self.image_item = Gtk.MenuItem(_("_Image"))
        # self.image_item.set_use_underline(True)
        # self.image_item.set_submenu(self.image_menu)
        # self.menu.append(self.image_item)
        #

        self.quotes_menu = Gtk.Menu()

        self.next_quote = Gtk.MenuItem(_("_Next"))
        self.next_quote.set_use_underline(True)
        self.next_quote.connect("activate", window.next_quote)
        self.quotes_menu.append(self.next_quote)

        self.prev_quote = Gtk.MenuItem(_("_Previous"))
        self.prev_quote.set_use_underline(True)
        self.prev_quote.connect("activate", window.prev_quote)
        self.quotes_menu.append(self.prev_quote)

        self.fast_forward_quote = Gtk.MenuItem(_("_Next, skipping forward history"))
        self.fast_forward_quote.set_use_underline(True)

        def _fast_forward_quote(widget):
            window.next_quote(widget, bypass_history=True)

        self.fast_forward_quote.connect("activate", _fast_forward_quote)
        self.quotes_menu.append(self.fast_forward_quote)

        self.quotes_menu.append(Gtk.SeparatorMenuItem.new())

        self.quotes_pause_resume = Gtk.MenuItem(_("Pause on current"))
        self.quotes_pause_resume.connect("activate", window.on_quotes_pause_resume)
        self.quotes_menu.append(self.quotes_pause_resume)

        self.quotes_menu.append(Gtk.SeparatorMenuItem.new())

        self.quote_favorite = Gtk.MenuItem(_("Save to Favorites"))
        self.quote_favorite.set_use_underline(True)
        self.quote_favorite.connect("activate", window.quote_save_to_favorites)
        self.quotes_menu.append(self.quote_favorite)

        self.quote_view_favs = Gtk.MenuItem(_("View Favorites..."))
        self.quote_view_favs.set_use_underline(True)
        self.quote_view_favs.connect("activate", window.quote_view_favorites)
        self.quotes_menu.append(self.quote_view_favs)

        self.quotes_menu.append(Gtk.SeparatorMenuItem.new())

        self.quote_clipboard = Gtk.MenuItem(_("Copy to Clipboard"))
        self.quote_clipboard.set_use_underline(True)
        self.quote_clipboard.connect("activate", window.quote_copy_to_clipboard)
        self.quotes_menu.append(self.quote_clipboard)

        self.view_quote = Gtk.MenuItem()
        self.view_quote.set_use_underline(True)
        self.view_quote.connect("activate", window.view_quote)
        self.quotes_menu.append(self.view_quote)

        self.google_quote_text = Gtk.MenuItem(_("Google Quote"))
        self.google_quote_text.set_use_underline(True)
        self.google_quote_text.connect("activate", window.google_quote_text)
        self.quotes_menu.append(self.google_quote_text)

        self.google_quote_author = Gtk.MenuItem(_("Google Author"))
        self.google_quote_author.set_use_underline(True)
        self.google_quote_author.connect("activate", window.google_quote_author)
        self.quotes_menu.append(self.google_quote_author)

        self.quotes_menu.append(Gtk.SeparatorMenuItem.new())

        self.quotes_preferences = Gtk.MenuItem(_("Preferences..."))
        self.quotes_preferences.set_use_underline(True)

        def _quotes_prefs(widget=None):
            window.preferences_dialog.ui.notebook.set_current_page(1)
            window.on_mnu_preferences_activate()

        self.quotes_preferences.connect("activate", _quotes_prefs)
        self.quotes_menu.append(self.quotes_preferences)

        self.quotes_disable = Gtk.MenuItem(_("Turn off"))
        self.quotes_disable.set_use_underline(True)
        self.quotes_disable.connect("activate", window.disable_quotes)
        self.quotes_menu.append(self.quotes_disable)

        self.quotes = Gtk.MenuItem(_("_Quote"))
        self.quotes.set_use_underline(True)
        self.quotes.set_submenu(self.quotes_menu)
        self.menu.append(self.quotes)

        self.menu.append(Gtk.SeparatorMenuItem.new())

        self.history = Gtk.CheckMenuItem(_("_History"))
        self.history.set_active(False)
        self.history.set_use_underline(True)
        self.history_handler_id = self.history.connect("toggled", window.show_hide_history)
        self.menu.append(self.history)

        self.selector = Gtk.CheckMenuItem(_("_Wallpaper Selector"))
        self.selector.set_active(False)
        self.selector.set_use_underline(True)
        self.selector_handler_id = self.selector.connect(
            "toggled", window.show_hide_wallpaper_selector
        )
        self.menu.append(self.selector)

        self.downloads = Gtk.CheckMenuItem(_("Recent _Downloads"))
        self.downloads.set_active(False)
        self.downloads.set_use_underline(True)
        self.downloads_handler_id = self.downloads.connect("toggled", window.show_hide_downloads)
        self.menu.append(self.downloads)

        if Util.check_variety_slideshow_present():
            self.menu.append(Gtk.SeparatorMenuItem.new())
            self.slideshow = Gtk.MenuItem(_("Start Slideshow"))
            self.slideshow.connect("activate", window.on_start_slideshow)
            self.menu.append(self.slideshow)
        else:
            logger.warning(
                "Variety Slideshow is not installed. This is an optional extension "
                "adding pan-and-zoom slideshows to Variety: see "
                "https://github.com/peterlevi/variety-slideshow for details"
            )

        self.menu.append(Gtk.SeparatorMenuItem.new())

        self.preferences = Gtk.MenuItem(_("Preferences..."))
        self.preferences.connect("activate", window.on_mnu_preferences_activate)
        self.menu.append(self.preferences)

        self.about = Gtk.MenuItem(_("About"))
        self.about.connect("activate", window.on_mnu_about_activate)
        self.menu.append(self.about)

        self.donate = Gtk.MenuItem(_("Donate"))
        self.donate.connect("activate", window.on_mnu_donate_activate)
        self.menu.append(self.donate)

        self.quit = Gtk.MenuItem(_("Quit"))
        self.quit.connect("activate", window.on_quit)
        self.menu.append(self.quit)

        self.menu.show_all()

    def create_indicator(self, window):
        logger.info("indicator backend: %s", _indicator_backend)
        self.indicator = None
        self.status_icon = None
        self.visible = True

        def right_click_event(icon, button, time):
            self.menu.popup(None, None, Gtk.StatusIcon.position_menu, self.status_icon, 0, time)

        def left_click_event(data):
            self.menu.popup(
                None,
                None,
                Gtk.StatusIcon.position_menu,
                self.status_icon,
                0,
                Gtk.get_current_event_time(),
            )

        def on_indicator_scroll_status_icon(status_icon, event):
            window.on_indicator_scroll(None, 1, event.direction)

        icon_path = varietyconfig.get_data_file("media", "variety-indicator.png")
        if use_appindicator:
            self.indicator = AppIndicator3.Indicator.new(
                "variety", "", AppIndicator3.IndicatorCategory.APPLICATION_STATUS
            )
            self.indicator.set_status(AppIndicator3.IndicatorStatus.ACTIVE)
            self.indicator.set_icon(icon_path)
            self.indicator.connect("scroll-event", window.on_indicator_scroll)
            self.indicator.set_menu(self.menu)
        else:
            self.status_icon = Gtk.StatusIcon.new_from_file(icon_path)
            self.status_icon.set_visible(True)
            self.status_icon.connect("activate", left_click_event)
            self.status_icon.connect("popup-menu", right_click_event)
            self.status_icon.connect("scroll-event", on_indicator_scroll_status_icon)

    def set_visible(self, visible):
        self.visible = visible
        if visible:
            if self.indicator:
                logger.info(lambda: "Showing indicator icon")
                self.indicator.set_status(AppIndicator3.IndicatorStatus.ACTIVE)
            if self.status_icon:
                logger.info(lambda: "Showing status icon")
                self.status_icon.set_visible(True)
        else:
            if self.indicator:
                logger.info(lambda: "Hiding indicator icon")
                self.indicator.set_status(AppIndicator3.IndicatorStatus.PASSIVE)
            if self.status_icon:
                logger.info(lambda: "Hiding status icon")
                self.status_icon.set_visible(False)

    def set_icon(self, icon):
        def set_from_theme_icon(name):
            if self.indicator:
                logger.info(lambda: "Showing indicator icon %s from GTK theme" % name)
                self.indicator.set_icon(name)
            if self.status_icon:
                logger.info(lambda: "Showing status icon %s from GTK theme" % name)
                self.status_icon.set_from_icon_name(name)

        if icon == "Light":
            if Gtk.IconTheme.get_default().has_icon(THEME_ICON_NAME):
                set_from_theme_icon(THEME_ICON_NAME)
                return
            else:
                icon_path = varietyconfig.get_data_file("media", "variety-indicator.png")
        elif icon == "Dark":
            if Gtk.IconTheme.get_default().has_icon(THEME_ICON_NAME_DARK):
                set_from_theme_icon(THEME_ICON_NAME_DARK)
                return
            else:
                icon_path = varietyconfig.get_data_file("media", "variety-indicator-dark.png")
        elif icon in ["1", "2", "3", "4"]:
            if Gtk.IconTheme.get_default().has_icon(THEME_ICON_NAME_NUM.format(icon)):
                set_from_theme_icon(THEME_ICON_NAME_NUM.format(icon))
                return
            else:
                icon_path = varietyconfig.get_data_file(
                    "media", "variety-indicator-num{}.png".format(icon)
                )
        elif icon and os.access(icon, os.R_OK) and Util.is_image(icon):
            icon_path = icon
        else:
            icon_path = varietyconfig.get_data_file("media", "variety-indicator.png")

        if self.indicator:
            logger.info(lambda: "Showing indicator icon image: " + icon_path)
            self.indicator.set_icon(icon_path)
        if self.status_icon:
            logger.info(lambda: "Showing status icon image: " + icon_path)
            self.status_icon.set_from_file(icon_path)

    def get_visible(self):
        return self.visible


def new_application_indicator(window):
    ind = Indicator(window)
    return ind, ind.indicator, ind.status_icon



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import random

from variety.plugins.downloaders.ImageSource import Throttling
from variety.plugins.downloaders.SimpleDownloader import SimpleDownloader
from variety.Util import Util, _

# Credits: Formerly using the data prepared by limhenry @ https://github.com/limhenry/earthview
DATA_URL = "https://earthview.withgoogle.com/_api/photos.json"

logger = logging.getLogger("variety")

random.seed()


class EarthviewDownloader(SimpleDownloader):
    DESCRIPTION = _("Google Earth View Wallpapers")
    ROOT_URL = "https://earthview.withgoogle.com/"

    @classmethod
    def get_info(cls):
        return {
            "name": "EarthviewDownloader",
            "description": EarthviewDownloader.DESCRIPTION,
            "author": "Peter Levi",
            "version": "0.1",
        }

    def get_description(self):
        return EarthviewDownloader.DESCRIPTION

    def get_source_type(self):
        return "earthview"

    def get_source_name(self):
        return "Earth View"

    def get_source_location(self):
        return self.ROOT_URL

    def fill_queue(self):
        queue = Util.fetch_json(DATA_URL)
        random.shuffle(queue)
        return queue

    def get_default_throttling(self):
        # throttle this source, as otherwise maps "overpower" all other types of images
        # with Variety's default settings, and we have no other way to control source "weights"
        return Throttling(max_downloads_per_hour=20, max_queue_fills_per_hour=None)

    def download_queue_item(self, item):
        item = Util.fetch_json("https://earthview.withgoogle.com/_api/" + item["slug"] + ".json")
        region = item["region"]
        filename = "{}{} (ID-{}).jpg".format(
            region + ", " if region and region != "-" else "", item["country"], item["id"]
        )
        origin_url = EarthviewDownloader.ROOT_URL + str(item["slug"])
        image_url = item["photoUrl"]
        if not image_url.startswith("http"):
            image_url = "https://" + image_url

        extra_metadata = {"description": item.get("name"), "author": item.get("attribution")}
        return self.save_locally(
            origin_url, image_url, local_filename=filename, extra_metadata=extra_metadata
        )



from typing import List

from variety.plugins.IDisplayModesPlugin import (
    DisplayMode,
    DisplayModeData,
    IDisplayModesPlugin,
    StaticDisplayMode,
)
from variety.Util import Util, _

IMAGEMAGICK_ZOOM = "-scale %Wx%H^ "
IMAGEMAGICK_FIT_WITH_BLACK = "-resize %Wx%H -size %Wx%H xc:black +swap -gravity center -composite"
IMAGEMAGICK_FIT_WITH_BLUR = (
    "-resize %Wx%H^ -gravity center -extent %Wx%H -scale 10% -blur 0x3 -resize 1000% -clone 0 "
    "-resize %Wx%H -size %Wx%H -gravity center -composite"
)


def _smart_fn(filename):
    try:
        image_w, image_h = Util.get_size(filename)
        primary_w, primary_h = Util.get_primary_display_size(hidpi_scaled=True)
        total_w, total_h = Util.get_multimonitor_display_size()
        if image_w * image_h * 10 < primary_w * primary_h:
            # image way smaller than primary monitor, tile it
            return DisplayModeData(set_wallpaper_param="wallpaper")
        else:
            image_ratio = image_w / image_h
            primary_ratio = primary_w / primary_h
            total_ratio = total_w / total_h
            if 2 * abs(image_ratio - primary_ratio) / (image_ratio + primary_ratio) < 0.2:
                # image ratio is close to primary screen ratio, zoom
                return DisplayModeData(set_wallpaper_param="zoom")
            elif 2 * abs(image_ratio - total_ratio) / (image_ratio + total_ratio) < 0.2:
                # image ratio is close to multimonitor total screen ratio, span it
                return DisplayModeData(set_wallpaper_param="spanned")
            else:
                # image ratio not close to screen ratio, fit with a blurred background
                cmd = IMAGEMAGICK_FIT_WITH_BLUR.replace("%W", str(primary_w)).replace(
                    "%H", str(primary_h)
                )
                return DisplayModeData(set_wallpaper_param="zoom", imagemagick_cmd=cmd)
    except:
        return DisplayModeData(set_wallpaper_param="zoom")


class ResizingDisplayModesPlugin(IDisplayModesPlugin):
    @classmethod
    def get_info(cls):
        return {
            "name": "ResizingDisplayModesPlugin",
            "description": "Display modes that use image resizing within Variety",
            "version": "1.0",
            "author": "Peter Levi",
        }

    def display_modes(self) -> List[DisplayMode]:
        return [
            DisplayMode(
                id="smart",
                title=_("Smart: Variety picks best mode based on image size. Recommended."),
                description=(
                    "Variety uses the fast OS-provided Zoom mode for images that are close to "
                    "screen proportions, uses 'Fit & pad with a blurred background' when the image "
                    "proportions are significantly different - e.g. portraits on a horizontal "
                    "screen, and uses the OS-provided tiling mode for very small images that would "
                    "look bad resized."
                ),
                fn=_smart_fn,
            ),
            StaticDisplayMode(
                id="zoom",
                title=_("Zoom to fill screen"),
                description=_(
                    "Image is zoomed in or out so that it fully fills your primary screen. "
                    "Some parts of the image will be cut out if its resolution is different "
                    "from the screen's. Slower than using native OS resizing options."
                ),
                set_wallpaper_param="zoom",
                imagemagick_cmd=IMAGEMAGICK_ZOOM,
            ),
            StaticDisplayMode(
                id="fill-with-black",
                title=_("Fit within screen, pad with black"),
                description=_(
                    "Image is zoomed in or out so that it fully fits within your primary screen. "
                    "The rest of the screen is filled with black. "
                    "Slower than using native OS resizing options."
                ),
                set_wallpaper_param="zoom",
                imagemagick_cmd=IMAGEMAGICK_FIT_WITH_BLACK,
            ),
            StaticDisplayMode(
                id="fill-with-blur",
                title=_("Fit within screen, pad with a blurred background. Slower."),
                description=_(
                    "Image is zoomed in or out so that it fully fits within your primary screen. "
                    "The rest of the screen is a filled with blurred version of the image. "
                ),
                set_wallpaper_param="zoom",
                imagemagick_cmd=IMAGEMAGICK_FIT_WITH_BLUR,
            ),
        ]

    def order(self):
        return 2000



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

"""Helpers for an Ubuntu application."""
import logging
import os

from .Builder import Builder
from .varietyconfig import get_data_file


def get_builder(builder_file_name):
    """Return a fully-instantiated Gtk.Builder instance from specified ui
    file

    :param builder_file_name: The name of the builder file, without extension.
        Assumed to be in the 'ui' directory under the data path.
    """
    # Look for the ui file that describes the user interface.
    ui_filename = get_data_file("ui", "%s.ui" % (builder_file_name,))
    if not os.path.exists(ui_filename):
        ui_filename = None

    builder = Builder()
    builder.set_translation_domain("variety")
    builder.add_from_file(ui_filename)
    return builder


# Owais Lone : To get quick access to icons and stuff.
def get_media_file(media_file_name):
    media_filename = get_data_file("media", "%s" % (media_file_name,))
    if not os.path.exists(media_filename):
        media_filename = None

    return "file:///" + media_filename


def get_help_uri(page=None):
    # help_uri from source tree - default language
    here = os.path.dirname(__file__)
    help_uri = os.path.abspath(os.path.join(here, "..", "help", "C"))

    if not os.path.exists(help_uri):
        # installed so use gnome help tree - user's language
        help_uri = "variety"

    # unspecified page is the index.page
    if page is not None:
        help_uri = "%s#%s" % (help_uri, page)

    return help_uri


def show_uri(parent, link):
    from gi.repository import Gtk  # pylint: disable=E0611

    screen = parent.get_screen()
    Gtk.show_uri(screen, link, Gtk.get_current_event_time())


def alias(alternative_function_name):
    """see http://www.drdobbs.com/web-development/184406073#l9"""

    def decorator(function):
        """attach alternative_function_name(s) to function"""
        if not hasattr(function, "aliases"):
            function.aliases = []
        function.aliases.append(alternative_function_name)
        return function

    return decorator



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import threading
import urllib.parse

from gi.repository import Gdk, Gtk  # pylint: disable=E0611

from variety.FlickrDownloader import FlickrDownloader
from variety.Options import Options
from variety.Util import Util, _, on_gtk
from variety_lib.helpers import get_builder

logger = logging.getLogger("variety")


class AddFlickrDialog(Gtk.Dialog):
    __gtype_name__ = "AddFlickrDialog"

    def __new__(cls):
        """Special static method that's automatically called by Python when 
        constructing a new instance of this class.
        
        Returns a fully instantiated AddFlickrDialog object.
        """
        builder = get_builder("AddFlickrDialog")
        new_object = builder.get_object("add_flickr_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called when we're finished initializing.

        finish_initalizing should be called after parsing the ui definition
        and creating a AddFlickrDialog object with it in order to
        finish initializing the start of the new AddFlickrDialog
        instance.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)
        self.edited_row = None

    def set_edited_row(self, edited_row):
        self.edited_row = edited_row

        location = edited_row[2]
        s = location.split(";")
        params = {}
        for x in s:
            if len(x) and x.find(":") > 0:
                k, v = x.split(":")
                params[k.lower()] = urllib.parse.unquote_plus(v)

        if "text" in params:
            self.ui.text.set_text(params["text"])

        if "tags" in params:
            self.ui.tags.set_text(params["tags"])

        if "user" in params:
            if "://" not in params["user"]:
                params["user"] = "https://" + params["user"]
            self.ui.user_url.set_text(params["user"])

        if "group" in params:
            if "://" not in params["group"]:
                params["group"] = "https://" + params["group"]
            self.ui.group_url.set_text(params["group"])

    def on_btn_ok_clicked(self, widget, data=None):
        """The user has elected to save the changes.

        Called before the dialog returns Gtk.ResponseType.OK from run().
        """
        threading.Timer(0, self.ok_thread).start()

    @on_gtk
    def show_spinner(self):
        self.ui.buttonbox.set_sensitive(False)
        self.ui.message.set_visible(True)
        self.ui.spinner.set_visible(True)
        self.ui.spinner.start()
        self.ui.error.set_label("")

    def ok_thread(self):
        search = ""

        if len(self.ui.tags.get_text().strip()):
            search += (
                "tags:"
                + ",".join(
                    [urllib.parse.quote_plus(t.strip()) for t in self.ui.tags.get_text().split(",")]
                )
                + ";"
            )

        if len(self.ui.text.get_text().strip()):
            search += "text:" + urllib.parse.quote_plus(self.ui.text.get_text().strip()) + ";"

        self.error = ""

        user_url = self.ui.user_url.get_text().strip()
        if len(user_url) > 0:
            self.show_spinner()
            u = FlickrDownloader.obtain_userid(user_url)
            if u[0]:
                search += (
                    "user:"
                    + self.ui.user_url.get_text().replace("http://", "").replace("https://", "")
                    + ";"
                )
                search += "user_id:" + u[2] + ";"
            else:
                self.error = self.error + "\n" + u[1]

        group_url = self.ui.group_url.get_text().strip()
        if len(group_url) > 0:
            self.show_spinner()
            g = FlickrDownloader.obtain_groupid(group_url)
            if g[0]:
                search += (
                    "group:"
                    + self.ui.group_url.get_text().replace("http://", "").replace("https://", "")
                    + ";"
                )
                search += "group_id:" + g[2]
            else:
                self.error = self.error + "\n" + g[1]

        if not len(self.error) and len(search) > 0:
            self.show_spinner()
            if FlickrDownloader.count_search_results(search) <= 0:
                self.error = _("No images found")

        def _stop_ui():
            self.ui.buttonbox.set_sensitive(True)
            self.ui.spinner.stop()
            self.ui.spinner.set_visible(False)
            self.ui.message.set_visible(False)

            if len(self.error) > 0:
                self.ui.error.set_label(self.error)
                for entry in (self.ui.text, self.ui.tags, self.ui.user_url, self.ui.group_url):
                    if entry.get_text().strip():
                        entry.grab_focus()
                        break
            else:
                if len(search):
                    self.parent.on_add_dialog_okay(
                        Options.SourceType.FLICKR, search, self.edited_row
                    )
                self.destroy()

        Util.add_mainloop_task(_stop_ui)

    def on_btn_cancel_clicked(self, widget, data=None):
        """The user has elected cancel changes.

        Called before the dialog returns Gtk.ResponseType.CANCEL for run()
        """
        self.destroy()


if __name__ == "__main__":
    dialog = AddFlickrDialog()
    dialog.show()
    Gtk.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
from os.path import basename, splitext

from variety.plugins.downloaders.SimpleDownloader import SimpleDownloader
from variety.Util import Util, _

DATA_URL = "https://www.nationalgeographic.co.uk/page-data/photo-of-day/page-data.json"

logger = logging.getLogger("variety")


class NationalGeographicDownloader(SimpleDownloader):
    DESCRIPTION = _("National Geographic's photo of the day")
    ROOT_URL = "https://www.nationalgeographic.co.uk/photo-of-day"

    @classmethod
    def get_info(cls):
        return {
            "name": "NationalGeographicDownloader",
            "description": NationalGeographicDownloader.DESCRIPTION,
            "author": "Eric Rösch",
            "version": "0.1",
        }

    def get_description(self):
        return NationalGeographicDownloader.DESCRIPTION

    def get_source_type(self):
        return "natgeo"

    def get_source_name(self):
        return "National Geographic"

    def get_source_location(self):
        return self.ROOT_URL

    def fill_queue(self):
        queue = Util.fetch_json(DATA_URL)
        images = queue["result"]["pageContext"]["node"]["data"]["content"]["images"]
        return images

    def download_queue_item(self, item):
        url = item["entity"]["mediaImage"]["url"]
        origin_url = NationalGeographicDownloader.ROOT_URL + "?image=" + splitext(basename(url))[0]
        image_url = "https://static.nationalgeographic.co.uk" + url
        author = item["entity"]["credit"]
        extra_metadata = {
            "author": author,
            "authorURL": "http://google.com/search?sourceid=variety&q=" + author,
            "headline": item["entity"]["mediaImage"]["alt"],
            "description": item["entity"]["caption"][3:-6],
        }
        return self.save_locally(origin_url, image_url, extra_metadata=extra_metadata)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

__all__ = ["project_path_not_found", "get_data_file", "get_data_path"]

try:
    from .variety_build_settings import __variety_data_directory__
except ImportError:
    # Variety's data directory. This is set by setup.py for permanent installations, but defaults to ../data
    # for easy development / running from source.
    __variety_data_directory__ = "../data"
__license__ = "GPL-3"
__version__ = "0.8.12"

import os


class project_path_not_found(Exception):
    """Raised when we can't find the project directory."""


def get_data_file(*path_segments):
    """Get the full path to a data file.

    Returns the path to a file underneath the data directory (as defined by
    `get_data_path`). Equivalent to os.path.join(get_data_path(),
    *path_segments).
    """
    return os.path.join(get_data_path(), *path_segments)


def get_data_path():
    """Retrieve variety data path

    This path is by default <variety_lib_path>/../data/ in trunk
    and /usr/share/variety in an installed version but this path
    is specified at installation time.
    """

    # Get pathname absolute or relative.
    path = os.path.join(os.path.dirname(__file__), __variety_data_directory__)

    abs_data_path = os.path.abspath(path)
    if not os.path.exists(abs_data_path):
        raise project_path_not_found

    return abs_data_path


def get_version():
    return __version__



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random
import re
import urllib.parse

from variety.plugins.downloaders.DefaultDownloader import DefaultDownloader
from variety.Util import Util

logger = logging.getLogger("variety")


class RedditDownloader(DefaultDownloader):
    def __init__(self, source, url):
        DefaultDownloader.__init__(self, source=source, config=url)

    @staticmethod
    def build_json_url(url):
        p = urllib.parse.urlparse(url)
        return (
            p.scheme
            + "://"
            + p.netloc
            + p.path
            + ".json"
            + "?"
            + p.query
            + ("&" if p.query else "")
            + "limit=100"
        )

    def fill_queue(self):
        logger.info(lambda: "Reddit URL: " + self.config)

        queue = []
        json_url = RedditDownloader.build_json_url(self.config)
        s = Util.fetch_json(json_url)
        for item in s["data"]["children"]:
            try:
                data = item["data"]
                image_url = data["url"]
                if re.match(r"^http(s)?://imgur\.com/\w+$", image_url):
                    image_url = image_url.replace("://", "://i.") + ".jpg"

                if image_url.lower().endswith((".jpg", ".jpeg", ".png")):
                    src_url = "https://www.reddit.com" + data["permalink"]
                    extra_metadata = {"sourceType": "reddit"}
                    if data["over_18"]:
                        extra_metadata["sfwRating"] = 0
                        if self.is_safe_mode_enabled():
                            continue
                    queue.append((src_url, image_url, extra_metadata))
            except Exception:
                logger.exception(lambda: "Could not process an item in the Reddit json result")

        random.shuffle(queue)
        return queue



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import json
from collections import defaultdict


class AttrDict(defaultdict):
    @staticmethod
    def converted(v):
        if isinstance(v, AttrDict):
            return v
        elif isinstance(v, dict):
            return AttrDict(v)
        elif isinstance(v, (list, tuple)):
            r = list(map(AttrDict.converted, v))
            return tuple(r) if isinstance(v, tuple) else r
        else:
            return v

    def merge(self, arg):
        if hasattr(arg, "items"):
            self.merge(arg.items())
        else:
            for k, v in arg:
                self[k] = AttrDict.converted(v)

    def asdict(self):
        return json.loads(json.dumps(self))

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(AttrDict)
        if len(args) == 1:
            self.merge(args[0])
        elif len(args) > 1:
            raise TypeError(
                "AttrDict expected at most 1 argument that is a map, got %i" % len(args)
            )
        self.merge(kwargs)

    def __setitem__(self, k, v):
        return super(AttrDict, self).__setitem__(k, AttrDict.converted(v))

    __getattr__ = defaultdict.__getitem__
    __setattr__ = __setitem__



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import random
import threading
import time

from variety.plugins.IQuoteSource import IQuoteSource
from variety.Util import _

logger = logging.getLogger("variety")


class QuotesEngine:
    def __init__(self, parent=None):
        self.parent = parent
        self.quote = None
        self.started = False
        self.running = False
        self.used = []

    def update_plugins(self):
        for p in self.parent.jumble.get_plugins(IQuoteSource):
            name = p["info"]["name"]
            if name in self.parent.options.quotes_disabled_sources:
                try:
                    p["plugin"].deactivate()
                except Exception:
                    logger.exception(lambda: "Error deactivating %s" % name)
            else:
                try:
                    p["plugin"].activate()
                except Exception:
                    logger.exception(lambda: "Error activating %s" % name)

        self.plugins = self.parent.jumble.get_plugins(IQuoteSource, active=True)

    def stop(self):
        self.running = False
        self.started = False
        self.update_plugins()

    def start(self):
        if self.started or not self.parent.options.quotes_enabled:
            return

        logger.info(lambda: "Starting QuotesEngine")

        self.update_plugins()

        self.prepared = []
        self.position = 0
        self.prepared_lock = threading.Lock()
        self.prepare_event = threading.Event()
        self.change_event = threading.Event()

        self.cache = {}

        self.started = True
        self.running = True

        self.last_change_time = time.time()
        self.last_error_notification_time = 0

        prep_thread = threading.Thread(target=self.prepare_thread)
        prep_thread.daemon = True
        prep_thread.start()

        change_thread = threading.Thread(target=self.regular_change_thread)
        change_thread.daemon = True
        change_thread.start()

    def quit(self):
        self.running = False
        self.prepare_event.set()

    def get_quote(self):
        return self.quote

    def has_previous(self):
        return self.position < len(self.used) - 1

    def prev_quote(self):
        self.last_change_time = time.time()
        self.position += 1
        if self.position >= len(self.used):
            if self.used:
                self.quote = self.choose_some_quote()
            self.used.append(self.quote)
        else:
            self.quote = self.used[self.position]
        return self.quote

    def bypass_history(self):
        self.position = 0

    def next_quote(self, bypass_history=False):
        self.last_change_time = time.time()
        if self.position > 0 and not bypass_history:
            self.position -= 1
            if self.position < len(self.used) - 1:
                self.quote = self.used[self.position]
            return self.quote
        else:
            if bypass_history:
                self.bypass_history()
            return self.change_quote()

    def choose_some_quote(self):
        with self.prepared_lock:
            if [x for x in self.prepared if x != self.quote]:
                self.quote = random.choice([x for x in self.prepared if x != self.quote])
            elif [x for x in self.used if x != self.quote]:
                self.quote = random.choice([x for x in self.used if x != self.quote])
            elif self.prepared:
                self.quote = random.choice(self.prepared)
            elif self.used:
                self.quote = random.choice(self.used)

            if self.quote in self.prepared:
                self.prepared.remove(self.quote)
                self.prepare_event.set()

            return self.quote

    def change_quote(self):
        self.last_change_time = time.time()

        self.choose_some_quote()

        self.used = self.used[self.position :]
        self.position = 0
        if self.quote:
            self.used.insert(0, self.quote)
        if len(self.used) > 200:
            self.used = self.used[:200]

        return self.quote

    def on_options_updated(self, clear_prepared=True):
        if clear_prepared:
            logger.info(lambda: "Quotes: clearing prepared and updating plugins")
            with self.prepared_lock:
                self.prepared = []
            self.update_plugins()
        self.prepare_event.set()
        self.change_event.set()

    def regular_change_thread(self):
        logger.info(lambda: "Quotes regular change thread running")

        while self.running:
            try:
                while (
                    not self.parent.options.quotes_change_enabled
                    or (time.time() - self.last_change_time)
                    < self.parent.options.quotes_change_interval
                ):
                    if not self.running:
                        return
                    now = time.time()
                    wait_more = self.parent.options.quotes_change_interval - max(
                        0, (now - self.last_change_time)
                    )
                    if self.parent.options.quotes_change_enabled:
                        self.change_event.wait(max(0, wait_more))
                    else:
                        self.change_event.wait()
                    self.change_event.clear()
                if not self.running:
                    return
                if not self.parent.options.quotes_change_enabled:
                    continue
                logger.info(lambda: "Quotes regular_change changes quote")
                self.last_change_time = time.time()
                self.parent.quote = self.change_quote()
                self.parent.refresh_texts()
            except Exception:
                logger.exception(lambda: "Exception in regular_change_thread")

    def prepare_thread(self):
        logger.info(lambda: "Quotes prepare thread running")

        while self.running:
            try:
                while (
                    self.running and self.parent.options.quotes_enabled and len(self.prepared) < 10
                ):
                    logger.info(
                        lambda: "Quotes prepared buffer contains %s quotes, fetching a quote"
                        % len(self.prepared)
                    )
                    quote = self.get_one_quote()
                    if quote:
                        with self.prepared_lock:
                            self.prepared.append(quote)
                        if self.parent.options.quotes_enabled and self.parent.quote is None:
                            self.parent.quote = self.change_quote()
                            self.parent.refresh_texts()

                    time.sleep(2)

                if not self.running:
                    return

            except Exception:
                logger.exception(lambda: "Error in quotes prepare thread:")

            self.prepare_event.wait()
            self.prepare_event.clear()

    def get_one_quote(self):
        keywords = []
        if self.parent.options.quotes_tags.strip():
            keywords = self.parent.options.quotes_tags.split(",")
        authors = []
        if self.parent.options.quotes_authors.strip():
            authors = self.parent.options.quotes_authors.split(",")

        category, search = ("random", "")
        if keywords or authors:
            category, search = random.choice(
                [("keyword", k) for k in keywords] + [("author", a) for a in authors]
            )

        plugins = list(self.plugins)
        if not self.parent.options.internet_enabled:
            plugins = [p for p in plugins if not p["plugin"].needs_internet()]
        if keywords or authors:
            plugins = [p for p in plugins if p["plugin"].supports_search()]

        if not plugins:
            self.parent.show_notification(
                _("No suitable quote plugins"),
                _(
                    "There are no quote plugins installed, "
                    "enabled and applicable for the current settings"
                ),
            )
            raise Exception("No quote plugins")

        error_plugins = []
        count_plugins = len(plugins)
        while self.running and self.parent.options.quotes_enabled:
            if not plugins:
                if (
                    time.time() - self.last_error_notification_time > 3600
                    and len(self.prepared) + len(self.used) < 5
                ):
                    self.last_error_notification_time = time.time()
                    if len(error_plugins) == count_plugins:
                        self.parent.show_notification(
                            _("Could not fetch quotes"),
                            _("Quotes services may be down, we will continue trying"),
                        )
                    else:
                        self.parent.show_notification(
                            _("Could not find quotes"),
                            _("Maybe you are searching for something very obscure?"),
                        )
                return None

            plugin = random.choice(plugins)
            plugin_name = plugin["info"]["name"]
            self.cache.setdefault(plugin_name, {"random": {}, "keyword": {}, "author": {}})
            self.cache[plugin_name][category].setdefault(search, {})
            cached = self.cache[plugin_name][category][search]

            if not cached:
                try:
                    if category == "random":
                        quotes = plugin["plugin"].get_random()
                    elif category == "keyword":
                        quotes = plugin["plugin"].get_for_keyword(search)
                    elif category == "author":
                        quotes = plugin["plugin"].get_for_author(search)
                    else:
                        raise RuntimeError("Unknown category")

                    if quotes:
                        for q in quotes:
                            if len(q["quote"]) < self.parent.options.quotes_max_length:
                                cached[q["quote"]] = q

                except Exception:
                    logger.exception(lambda: "Exception in quote plugin")
                    plugins.remove(plugin)
                    error_plugins.append(plugin)
                    continue

            if not cached:
                logger.warning(lambda: "No quotes for '%s' for plugin %s" % (search, plugin_name))
                plugins.remove(plugin)
                continue

            quote = random.choice(list(cached.values()))
            del cached[quote["quote"]]
            if not cached:
                del cached

            return quote



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import abc
import collections
import logging
import time
from datetime import timedelta

from variety.plugins.IVarietyPlugin import IVarietyPlugin

logger = logging.getLogger("variety")


Throttling = collections.namedtuple(
    "Throttling", ["max_downloads_per_hour", "max_queue_fills_per_hour"]
)


class ImageSource(IVarietyPlugin, metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()
        self._last_download_times = []
        self._last_queue_fill_times = []
        self.variety = None

    def set_variety(self, variety):
        """
        Sets the VarietyWindow instance. This is called after Jumble creates the instance, before
        it is actually used.
        :param variety: the instance of VarietyWindow
        """
        self.variety = variety

    def get_variety(self):
        """
        Returns the VarietyWindow instance, if set via set_variety(), or None.
        This is available before the source is actually used.
        :return the instance of VarietyWindow
        """
        return self.variety

    @abc.abstractmethod
    def get_source_type(self):
        """
        Returns a key for this source that will be used to identify it in
        configuration files, and will be saved in image metadata under Xmp.variety.sourceType.
        Variety will use this source type to find which plugin can handle a particular
        saved image source configuration, so this should not collide between different image
        source plugins.
        :return: source type, e.g. "flickr", "unsplash", etc.
        """
        pass

    def get_source_name(self):
        """
        Returns the value that will go into Xmp.variety.sourceName in image metadata.
        This will also be shown in UI.
        This could be the name of the service that is used to fetch images, e.g. Flickr.
        Default implementation is to return source_type, with uppercase first letter.
        :return: source name, e.g. "Flickr", "Unsplash", etc.
        """
        source_type = self.get_source_type()
        return source_type[0].upper() + source_type[1:]

    def needs_internet(self):
        """
        Does this configurable image source need internet in order to fetch new images?
        :return: True or False
        """
        return True

    def on_image_set_as_wallpaper(self, img, meta):
        """
        Called when a wallpaper downloaded from this source was used as a wallpaper.
        This can be used to call back the image provider for stats purposes.
        :param img path to the image file
        :param meta image metadata
        """
        pass

    def on_image_favorited(self, img, meta):
        """
        Called when a wallpaper downloaded from this source was copied or moved to favorites.
        This can be used to call back the image provider for stats purposes.
        :param img path to the image file
        :param meta image metadata
        """
        pass

    def get_default_throttling(self):
        """
        Throttling serves to avoid overloading servers when multiple Variety users use the source 
        simultaneously. It is normally controlled via a remote config, but defaults should be 
        provided for the cases when the remote config is not set or cannot be fetched.
        All downloaders for the same source are throttled together.
        :return: a Throttling namedtuple
        """
        return Throttling(max_downloads_per_hour=None, max_queue_fills_per_hour=None)

    def get_server_options_key(self):
        """
        Key under the the server-side throttling options where the configs for this source reside.
        By default it is the same as the source type.
        :return: key in remote server options for this source, e.g. "unsplash_v2"
        """
        return self.get_source_type()

    def get_server_options(self):
        """
        Returns the server options for this source.
        The default implementation reads from get_variety().server_options, i.e. it uses Variety's central
        serverside options, but you could override this method to read from elsewhere.
        :return: remotely-configured options for this image source
        """
        return self.get_variety().server_options[self.get_server_options_key()]

    def get_throttling(self):
        """
        Returns the actual throttling, taking remote configuration into account (if available)
        """
        defaults = self.get_default_throttling()

        max_downloads_per_hour, max_queue_fills_per_hour = defaults
        name = self.get_source_name()

        try:
            logger.info(lambda: "{}: parsing serverside options".format(name))
            options = self.get_server_options()
            logger.info(
                lambda: "{} serverside options: {}".format(self.get_source_name(), str(options))
            )
        except Exception:
            logger.info(
                lambda: "Could not parse {} serverside options, using defaults {}, {}".format(
                    name, max_downloads_per_hour, max_queue_fills_per_hour
                )
            )
            return defaults

        try:
            max_downloads_per_hour = int(options["max_downloads_per_hour"])
        except Exception:
            pass

        try:
            max_queue_fills_per_hour = int(options["max_queue_fills_per_hour"])
        except Exception:
            pass

        return Throttling(max_downloads_per_hour, max_queue_fills_per_hour)

    def _count_last_hour_downloads(self):
        now = time.time()
        self._last_download_times = [t for t in self._last_download_times if now - t < 3600]
        return len(self._last_download_times)

    def is_download_allowed(self):
        max_downloads_per_hour, _ = self.get_throttling()
        return (
            max_downloads_per_hour is None
            or self._count_last_hour_downloads() < max_downloads_per_hour
        )

    def register_download(self):
        self._last_download_times.append(time.time())

    def _count_last_hour_queue_fills(self):
        now = time.time()
        self._last_queue_fill_times = [t for t in self._last_queue_fill_times if now - t < 3600]
        return len(self._last_queue_fill_times)

    def is_fill_queue_allowed(self):
        _, max_queue_fills_per_hour = self.get_throttling()
        return (
            max_queue_fills_per_hour is None
            or self._count_last_hour_queue_fills() < max_queue_fills_per_hour
        )

    def register_fill_queue(self):
        self._last_queue_fill_times.append(time.time())



from variety.Util import Util, _


def _os(filename):
    return "os", None


def _zoom(filename):
    w, h = Util.get_primary_display_size()
    return "zoom", "-scale %dx%d^ " % (w, h)


def _fill_with_black(filename):
    w, h = Util.get_primary_display_size()
    return (
        "zoom",
        "-resize %dx%d\> -size %dx%d xc:black +swap -gravity center -composite" % (w, h, w, h),
    )


def _fill_with_blur(filename):
    w, h = Util.get_primary_display_size()
    return (
        "zoom",
        "-resize %dx%d^ -gravity center -extent %dx%d -blur 0x10 -clone 0 -resize %dx%d -size %dx%d -gravity center -composite"
        % (w, h, w, h, w, h, w, h),
    )


DISPLAY_MODES = [
    {
        "id": "os",
        "title": _("Controlled by OS settings, not by Variety. Fastest option."),
        "description": _(
            "Display mode is controlled by your OS Appearance settings and by what is "
            "specified in set_wallpaper script for your desktop environment. "
        ),
        "fn": _os,
    },
    {
        "id": "zoom",
        "title": _("Zoom"),
        "description": _(
            "Image is zoomed in or out so that it fully fills your primary screen. "
            "Some parts of the image will be cut out if its resolution is different "
            "from the screen's. "
        ),
        "fn": _zoom,
    },
    {
        "id": "fill-with-black",
        "title": _("Fill with black"),
        "description": _(
            "Image is zoomed in or out so that it fully fits within your primary screen. "
            "The rest of the screen is filled with black. "
        ),
        "fn": _fill_with_black,
    },
    {
        "id": "fill-with-blur",
        "title": _("Fill with blur. Slow."),
        "description": _(
            "Image is zoomed in or out so that it fully fits within your primary screen. "
            "The rest of the screen is a filled with blurred version of the image. "
        ),
        "fn": _fill_with_blur,
    },
    {
        "id": "smart-with-black",
        "title": _("Pick dynamically between Zoom and Fill with black"),
        "description": _(
            "Variety picks a good mode depending on image size. "
            "Images that are close to the screen proportions use 'Zoom'. "
            "Those that are vastly different use 'Fill with black'. "
        ),
        "fn": _fill_with_black,  # TODO implement
    },
    {
        "id": "smart-with-blur",
        "title": _("Pick dynamically between Zoom and Fill with blur. Slow."),
        "description": _(
            "Variety picks a good mode depending on image size. "
            "Images that are close to the screen proportions use 'Zoom'. "
            "Those that are vastly different use 'Fill with blur'. "
        ),
        "fn": _fill_with_blur,  # TODO implement
    },
]



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
import logging
import random
import time

from variety.plugins.downloaders.ImageSource import Throttling
from variety.plugins.downloaders.SimpleDownloader import SimpleDownloader
from variety.Util import Util, _

logger = logging.getLogger("variety")

random.seed()


class UnsplashDownloader(SimpleDownloader):
    DESCRIPTION = _("High-resolution photos from Unsplash.com")

    API_KEY = "37ee4be36ffd27e79decd4ef382d4949"
    HASH = (
        b"AwBXAAFSUQtSAAUGBQQED11dBAdRDQMFVQwCVgAOBQwCDwMDDAYDBARTAgIGAVdVCQdRBQUCU1NV\n"
        b"DARSAQgBWA==\n"
    )
    UTM_PARAMS = "?utm_source=Variety+Wallpaper+Changer&utm_medium=referral"

    rate_limiting_started_time = 0

    @classmethod
    def get_info(cls):
        return {
            "name": "UnsplashDownloader",
            "description": UnsplashDownloader.DESCRIPTION,
            "author": "Peter Levi",
            "version": "0.1",
        }

    def get_source_type(self):
        return "unsplash"

    def get_description(self):
        return UnsplashDownloader.DESCRIPTION

    def get_source_name(self):
        return "Unsplash.com"

    def get_source_location(self):
        return "https://unsplash.com"

    def get_folder_name(self):
        return "Unsplash"

    def get_server_options_key(self):
        return "unsplash_v2"

    def get_default_throttling(self):
        return Throttling(max_downloads_per_hour=10, max_queue_fills_per_hour=1)

    def get_unsplash_api_url(self):
        return "https://api.unsplash.com/photos/random?count=30&client_id={}{}".format(
            Util.unxor(UnsplashDownloader.HASH, UnsplashDownloader.API_KEY),
            "&orientation=landscape" if self.get_variety().options.use_landscape_enabled else "",
        )

    def fill_queue(self):
        if time.time() - UnsplashDownloader.rate_limiting_started_time < 3600:
            logger.info(
                lambda: "Unsplash queue empty, but rate limit reached, will try again later"
            )
            return []

        url = self.get_unsplash_api_url()
        logger.info(lambda: "Filling Unsplash queue from " + url)

        r = Util.request(url)
        if int(r.headers.get("X-Ratelimit-Remaining", 1000000)) < 1000:
            UnsplashDownloader.rate_limiting_started_time = time.time()

        queue = []
        for item in r.json():
            try:
                width = item["width"]
                height = item["height"]
                if self.is_size_inadequate(width, height):
                    continue

                image_url = item["urls"]["full"] + "&w={}".format(
                    max(1980, int(Util.get_primary_display_size()[0] * 1.2))
                )
                origin_url = item["links"]["html"] + UnsplashDownloader.UTM_PARAMS
                
                if "plus.unsplash.com/" in image_url:  # exclude watermarked wallpapers
                    continue

                extra_metadata = {
                    "sourceType": "unsplash",
                    "sfwRating": 100,
                    "author": item["user"]["name"],
                    "authorURL": item["user"]["links"]["html"] + UnsplashDownloader.UTM_PARAMS,
                    "description": item.get("description", item.get("alt_description")),
                    "keywords": [
                        cat
                        for cat, approval in item.get("topic_submissions", {}).items()
                        if approval.get("status") == "approved"
                    ],
                    "extraData": {
                        "unsplashDownloadLocation": item["links"]["download_location"],
                        "unsplashDownloadReported": False,
                    },
                }

                queue.append((origin_url, image_url, extra_metadata))
            except:
                logger.exception(lambda: "Could not process an item from Unsplash")
                raise

        random.shuffle(queue)
        return queue

    def on_image_set_as_wallpaper(self, img, meta):
        extraData = meta.get("extraData", None)
        if not extraData:
            return

        download_loc = extraData.get("unsplashDownloadLocation")
        reported = extraData.get("unsplashDownloadReported")
        if download_loc and not reported:
            url = "{}?client_id={}".format(
                download_loc, Util.unxor(UnsplashDownloader.HASH, UnsplashDownloader.API_KEY)
            )
            Util.fetch(url)
            meta["extraData"]["unsplashDownloadReported"] = True
            Util.write_metadata(img, meta)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (c) 2012, Peter Levi <peterlevi@peterlevi.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import logging
import os
import threading
import time

from gi.repository import Gdk, GdkPixbuf, GObject, Gtk

from variety.profile import get_profile_wm_class
from variety.Util import Util, on_gtk

logger = logging.getLogger("variety")


class ThumbsWindow(Gtk.Window):
    __gsignals__ = {"clicked": (GObject.SIGNAL_RUN_FIRST, None, (str, Gtk.Widget, object))}

    LEFT = 1
    RIGHT = 2
    BOTTOM = 3
    TOP = 4

    def __init__(self, screen=None, position=BOTTOM, breadth=120):
        logger.debug(lambda: "Creating thumb window %s, %d" % (str(self), time.time()))
        super(ThumbsWindow, self).__init__()

        self.set_wmclass(get_profile_wm_class(), get_profile_wm_class())

        self.running = True

        self.set_decorated(False)
        self.set_accept_focus(False)

        self.screen = screen if screen else Gdk.Screen.get_default()
        self.monitor_area = self.screen.get_monitor_workarea(self.screen.get_primary_monitor())

        self.position = position
        self.breadth = int(
            breadth
            * (
                1
                if self.is_horizontal()
                else float(self.monitor_area.width) / self.monitor_area.height
            )
        )

        self.box = Gtk.HBox(False, 0) if self.is_horizontal() else Gtk.VBox(False, 0)

        self.scroll = Gtk.ScrolledWindow()
        self.scroll.add_with_viewport(self.box)
        if self.is_horizontal():
            self.scroll.set_min_content_height(self.breadth)
            self.scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER)
        else:
            self.scroll.set_min_content_width(self.breadth)
            self.scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self.mouse_in = False
        self.mouse_position = None
        self.autoscroll_event = threading.Event()
        self.scrolling_paused = False
        self.scrolling_locked = False

        def mouse_enter(widget, event, data=None):
            self.mouse_in = True
            self.previous_speed = 0
            self.autoscroll_event.set()

        def mouse_motion(widget, event, data=None):
            self.mouse_position = (event.x, event.y)
            if not self.scrolling_locked:
                self.scrolling_paused = False

        def mouse_leave(widget, event, data=None):
            self.mouse_in = False
            self.mouse_position = None
            self.previous_speed = 0
            self.autoscroll_event.clear()

        eventbox = Gtk.EventBox()
        eventbox.set_visible(True)
        eventbox.add(self.scroll)
        eventbox.set_events(
            Gdk.EventMask.ENTER_NOTIFY_MASK
            | Gdk.EventMask.LEAVE_NOTIFY_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
        )
        eventbox.connect("enter-notify-event", mouse_enter)
        eventbox.connect("leave-notify-event", mouse_leave)
        eventbox.connect("motion-notify-event", mouse_motion)

        self.add(eventbox)

        self.image_count = 0

        self.active_file = None
        self.active_position = None
        self.mark = None
        self.marked_info = None

        self.all = []

    def pause_scrolling(self):
        self.previous_speed = 0
        self.scrolling_paused = True
        self.scrolling_locked = True

    def resume_scrolling(self):
        self.scrolling_locked = False

    def is_horizontal(self):
        return self.position == ThumbsWindow.TOP or self.position == ThumbsWindow.BOTTOM

    def pin(self, widget=None):
        self.pinned = True

    @on_gtk
    def start(self, images):
        self.images = images

        self._show()

        thumbs_thread = threading.Thread(target=self._thumbs_thread)
        thumbs_thread.daemon = True
        thumbs_thread.start()

        autoscroll_thread = threading.Thread(target=self._autoscroll_thread)
        autoscroll_thread.daemon = True
        autoscroll_thread.start()

    def _calc_start_position(self):
        area = self.monitor_area
        if self.position == ThumbsWindow.BOTTOM:
            return area.x + area.width // 2, area.y + area.height - self.breadth
        elif self.position == ThumbsWindow.TOP:
            return area.x + area.width // 2, area.y
        elif self.position == ThumbsWindow.LEFT:
            return area.x, area.y + area.height // 2
        elif self.position == ThumbsWindow.RIGHT:
            return area.x + area.width - self.breadth, area.height // 2
        else:
            raise Exception("Unsupported thumbs position: " + str(self.position))

    def _show(self):
        self.set_default_size(1, 1)
        logger.debug(lambda: "Showing thumb window %s, %d" % (str(self), time.time()))
        self.move(*self._calc_start_position())
        self.show_all()

    def _thumbs_thread(self):
        logger.debug(lambda: "Starting thumb thread %s, %d" % (str(self), time.time()))
        try:
            self.total_width = 0

            for i, file in enumerate(self.images):
                if not self.running:
                    Util.add_mainloop_task(self.destroy)
                    return

                self.add_image(file, at_front=False)
                self.image_count = i

        except Exception:
            logger.exception(lambda: "Error while creating thumbs:")

    def add_image(self, file, at_front=False):
        try:
            if self.is_horizontal():
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(file, 10000, self.breadth)
            else:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(file, self.breadth, 10000)
        except Exception:
            logger.warning(
                lambda: "Could not create thumbnail for file %s. File may be missing or invalid."
                % file
            )
            pixbuf = None

        def _go():
            image_size = (
                0
                if not pixbuf
                else pixbuf.get_width()
                if self.is_horizontal()
                else pixbuf.get_height()
            )

            thumb = Gtk.Image()
            if pixbuf:
                thumb.set_from_pixbuf(pixbuf)
            thumb.set_visible(True)

            overlay = Gtk.Overlay()
            overlay.add(thumb)
            overlay.set_visible(True)

            eventbox = Gtk.EventBox()
            eventbox.set_visible(True)

            def click(widget, event, file=file):
                self.emit("clicked", file, widget, event)

            eventbox.connect("button-release-event", click)
            eventbox.add(overlay)

            image_info = {
                "file": file,
                "eventbox": eventbox,
                "thumb": thumb,
                "size": image_size,
                "overlay": overlay,
            }
            if at_front:
                image_info["start"] = 0
                for info in self.all:
                    info["start"] += image_size
                self.all.insert(0, image_info)
                position = 0
            else:
                image_info["start"] = self.total_width
                self.all.append(image_info)
                position = len(self.all) - 1

            self.total_width += image_size

            adj = (
                self.scroll.get_hadjustment()
                if self.is_horizontal()
                else self.scroll.get_vadjustment()
            )
            scrollbar_at_start = not adj or adj.get_value() <= adj.get_lower() + 20

            self.box.pack_start(eventbox, False, False, 0)

            if at_front:
                self.box.reorder_child(eventbox, 0)
                # get adj again - we just added at front, scrollbar might have appeared
                adj = (
                    self.scroll.get_hadjustment()
                    if self.is_horizontal()
                    else self.scroll.get_vadjustment()
                )
                if adj:
                    if scrollbar_at_start:
                        adj.set_value(adj.get_lower())
                    else:
                        adj.set_value(adj.get_value() + image_size)

            self.update_size()

            if file == self.active_file or position == self.active_position:
                self.mark_active(self.active_file, self.active_position)

        Util.add_mainloop_task(_go)

    def _window_length(self):
        if self.is_horizontal():
            return self.monitor_area.width
        else:
            return self.monitor_area.height

    def _calc_position(self):
        area = self.monitor_area
        if self.position == ThumbsWindow.BOTTOM:
            return (
                area.x + max(0, (area.width - self.total_width) // 2),
                area.y + area.height - self.breadth,
            )
        elif self.position == ThumbsWindow.TOP:
            return (area.x + max(0, (area.width - self.total_width) // 2), area.y)
        elif self.position == ThumbsWindow.LEFT:
            return (area.x, area.y + max(0, (area.height - self.total_width) // 2))
        elif self.position == ThumbsWindow.RIGHT:
            return (
                area.x + area.width - self.breadth,
                area.y + max(0, (area.height - self.total_width) // 2),
            )
        else:
            raise Exception("Unsupported thumbs position: " + str(self.position))

    def update_size(self):
        if self.total_width < self._window_length() + 1000:
            self.move(*self._calc_position())
            if self.is_horizontal():
                self.scroll.set_min_content_width(min(self.total_width, self.monitor_area.width))
            else:
                self.scroll.set_min_content_height(min(self.total_width, self.monitor_area.height))

    # TODO this method is buggy when width < screen and scrollbar not shown - a blank space remains
    @on_gtk
    def remove_image(self, image):
        for info in self.all:
            if info["file"] == image:
                eventbox = info["eventbox"]
                thumb = info["thumb"]
                self.box.remove(eventbox)
                eventbox.destroy()
                thumb.destroy()
                self.total_width -= info["size"]
                self.update_size()

        self.all = [info for info in self.all if info["file"] != image]

    def mark_active(self, file=None, position=None):
        def _mark():
            logger.debug(lambda: "Marking file %s, position %s" % (file, str(position)))

            self.active_file = file
            self.active_position = position

            pos = position
            if self.active_file:
                try:
                    pos = [info["file"] for info in self.all].index(self.active_file)
                except Exception:
                    pass

            if self.mark:
                if self.marked_info:
                    self.marked_info["overlay"].remove(self.mark)
                    self.marked_info = None
                self.mark.destroy()
                self.mark = None

            if pos is not None and len(self.all) > pos:
                self.marked_info = self.all[pos]

                image_size = self.marked_info["size"]

                self.mark = Gtk.DrawingArea()
                if self.is_horizontal():
                    self.mark.set_size_request(image_size, 5)
                    self.mark.set_valign(Gtk.Align.START)
                    self.mark.set_halign(Gtk.Align.CENTER)
                else:
                    self.mark.set_size_request(5, image_size)
                    self.mark.set_valign(Gtk.Align.CENTER)
                    self.mark.set_halign(Gtk.Align.START)

                def _draw_callback(widget, cr):
                    if self.is_horizontal():
                        cr.rectangle(0, 0, image_size, 5)
                    else:
                        cr.rectangle(0, 0, 5, image_size)
                    cr.set_source_rgba(255.0 / 255, 105.0 / 255, 44.0 / 255)
                    cr.fill()
                    return False

                self.mark.connect("draw", _draw_callback)
                self.mark.set_visible(True)

                self.marked_info["overlay"].add_overlay(self.mark)

        GObject.idle_add(_mark)

    def fits_in_screen(self, with_reserve=0):
        if self.is_horizontal():
            return self.total_width < self.monitor_area.width + with_reserve

    def destroy(self, widget=False):
        logger.debug(lambda: "Destroying thumb window %s, %d" % (str(self), time.time()))
        self.running = False
        self.autoscroll_event.set()
        super(ThumbsWindow, self).destroy()

    def autoscroll_step(self, adj, total_size, current):
        if not adj:
            return

        if not hasattr(self, "previous_speed"):
            self.previous_speed = 0

        left_limit = total_size / 5
        right_limit = 4 * total_size / 5

        if current <= left_limit and adj.get_value() > adj.get_lower():
            speed = 30 * (left_limit - current) ** 3 / left_limit**3
            if adj.get_value() < adj.get_lower() + 800:
                speed = speed * (adj.get_value() - adj.get_lower()) / 800
            speed = min(speed, self.previous_speed + 0.1)
            self.previous_speed = speed
            adj.set_value(max(adj.get_lower(), adj.get_value() - speed))
        elif current >= right_limit and adj.get_value() < adj.get_upper():
            speed = 30 * (current - right_limit) ** 3 / (total_size - right_limit) ** 3
            if adj.get_value() > adj.get_upper() - adj.get_page_size() - 800:
                speed = speed * (adj.get_upper() - adj.get_page_size() - adj.get_value()) / 800
            speed = min(speed, self.previous_speed + 0.1)
            self.previous_speed = speed
            adj.set_value(min(adj.get_upper(), adj.get_value() + speed))

    def _autoscroll_thread(self):
        last_update = time.time()
        while self.running:
            while not self.mouse_in:
                if not self.running:
                    return
                self.autoscroll_event.wait(10)

            time.sleep(max(0, last_update + 0.005 - time.time()))

            if not self.mouse_position or self.scrolling_paused:
                continue

            x = self.mouse_position[0]
            y = self.mouse_position[1]

            def _go():
                if self.is_horizontal() and y > 0:
                    self.autoscroll_step(
                        self.scroll.get_hadjustment(), self.scroll.get_min_content_width(), x
                    )
                elif not self.is_horizontal() and x > 0:
                    self.autoscroll_step(
                        self.scroll.get_vadjustment(), self.scroll.get_min_content_height(), y
                    )

            Util.add_mainloop_task(_go)

            last_update = time.time()


if __name__ == "__main__":
    images = []
    dir = "/usr/share/backgrounds"
    for f in os.listdir(dir):
        file = os.path.join(dir, f)
        if os.path.isfile(file) and file.endswith(".jpg"):
            images.append(file)

    print(images)

    win = ThumbsWindow()
    win.connect("delete-event", Gtk.main_quit)

    print("starting")
    win.start(images)
    print("gtk main")
    Gtk.main()
